Package: s7contract
Title: 'Go'-Like Interfaces and 'Rust'-Like Traits with 'S7'
Version: 0.1.0
Authors@R: c(
    person(given = "Sounkou Mahamane", family = "Toure",
           email = "sounkoutoure@gmail.com", role = c("aut", "cre"))
    )
Description: Contract helpers built with 'S7' for expressing runtime protocols
    around ordinary 'S7' dispatch. Structural interfaces describe small sets of
    required 'S7' generics, while explicit traits record registered
    implementations with optional default methods and associated metadata.
    Optional runtime checks can validate argument and return specifications in
    contract-scoped evaluation.
License: GPL (>= 3)
Encoding: UTF-8
RoxygenNote: 7.3.3
Depends: R (>= 4.3.0)
Imports: S7
Suggests: knitr, rmarkdown, tinytest
VignetteBuilder: knitr
URL: https://github.com/sounkou-bioinfo/s7contract,
        https://sounkou-bioinfo.github.io/s7contract/
BugReports: https://github.com/sounkou-bioinfo/s7contract/issues
NeedsCompilation: no
Packaged: 2026-05-04 17:03:53 UTC; sounkoutoure
Author: Sounkou Mahamane Toure [aut, cre]
Maintainer: Sounkou Mahamane Toure <sounkoutoure@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-07 16:10:08 UTC
Built: R 4.5.2; ; 2026-05-07 21:02:47 UTC; unix

# Generated by roxygen2: do not edit by hand

export("%::%")
export(as_interface)
export(assert_implements)
export(assert_trait)
export(has_trait)
export(impl_trait)
export(implements)
export(interface_report)
export(interface_requirement)
export(interface_requirements)
export(missing_requirements)
export(new_interface)
export(new_trait)
export(trait_assoc_const)
export(trait_assoc_type)
export(trait_call)
export(trait_method)
export(trait_methods)
export(trait_report)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup--------------------------------------------------------------------
library(S7)
library(s7contract)

## ----toy-container------------------------------------------------------------
MiniSummarizedExperiment <- new_class(
  "MiniSummarizedExperiment",
  properties = list(
    assays = class_list,
    row_data = class_data.frame,
    col_data = class_data.frame
  ),
  validator = function(self) {
    if (length(self@assays) == 0) {
      return("@assays must contain at least one matrix")
    }

    dims <- lapply(self@assays, dim)
    if (any(vapply(dims, is.null, logical(1)))) {
      return("every assay must be matrix-like")
    }

    first_dim <- dims[[1]]
    same_dim <- vapply(dims, identical, logical(1), first_dim)
    if (!all(same_dim)) {
      return("all assays must have the same dimensions")
    }

    if (nrow(self@row_data) != first_dim[[1]]) {
      return("@row_data must have one row per assay feature")
    }
    if (nrow(self@col_data) != first_dim[[2]]) {
      return("@col_data must have one row per assay sample")
    }
  }
)

counts <- matrix(
  c(10, 0, 3, 4, 12, 8),
  nrow = 3,
  dimnames = list(c("geneA", "geneB", "geneC"), c("sample1", "sample2"))
)

mini <- MiniSummarizedExperiment(
  assays = list(counts = counts, logcounts = log1p(counts)),
  row_data = data.frame(gc = c(0.42, 0.51, 0.37), row.names = rownames(counts)),
  col_data = data.frame(condition = c("control", "treated"), row.names = colnames(counts))
)

## ----assay-operations---------------------------------------------------------
assay_names <- new_generic("assay_names", "x")
feature_names <- new_generic("feature_names", "x")
sample_names <- new_generic("sample_names", "x")
assay_matrix <- new_generic("assay_matrix", "x")

method(assay_names, MiniSummarizedExperiment) <- function(x) names(x@assays)
method(feature_names, MiniSummarizedExperiment) <- function(x) rownames(x@assays[[1]])
method(sample_names, MiniSummarizedExperiment) <- function(x) colnames(x@assays[[1]])
method(assay_matrix, MiniSummarizedExperiment) <- function(x, name = assay_names(x)[[1]]) {
  x@assays[[name]]
}

assay_names(mini)
sample_names(mini)
assay_matrix(mini, "counts")[, "sample1"]

## ----assay-consumer-----------------------------------------------------------
LibrarySizeInput <- new_interface(
  "LibrarySizeInput",
  generics = list(assay_matrix = assay_matrix)
)

library_size <- function(x, assay = "counts") {
  assert_implements(x, LibrarySizeInput)
  mat <- assay_matrix(x, assay)
  colSums(mat)
}

implements(mini, LibrarySizeInput)
library_size(mini)

## ----assay-mock---------------------------------------------------------------
MockAssays <- new_class("MockAssays", properties = list(assays = class_list))

method(assay_matrix, MockAssays) <- function(x, name = "counts") {
  x@assays[[name]]
}

mock_counts <- matrix(
  c(1, 2, 3, 4),
  nrow = 2,
  dimnames = list(c("geneA", "geneB"), c("sampleA", "sampleB"))
)
mock <- MockAssays(assays = list(counts = mock_counts))

implements(mock, LibrarySizeInput)
library_size(mock)

## ----assay-trait, message = FALSE, warning = FALSE----------------------------
ExperimentLike <- new_trait(
  "ExperimentLike",
  methods = list(
    assay_names = trait_method(assay_names),
    feature_names = trait_method(feature_names),
    sample_names = trait_method(sample_names),
    assay_matrix = trait_method(assay_matrix)
  ),
  assoc_consts = c("ASSAY_ORIENTATION")
)

impl_trait(
  ExperimentLike,
  MiniSummarizedExperiment,
  methods = list(
    assay_names = function(x) names(x@assays),
    feature_names = function(x) rownames(x@assays[[1]]),
    sample_names = function(x) colnames(x@assays[[1]]),
    assay_matrix = function(x, name = assay_names(x)[[1]]) x@assays[[name]]
  ),
  assoc_consts = list(ASSAY_ORIENTATION = "features_by_samples"),
  replace = TRUE
)

has_trait(mini, ExperimentLike)
trait_assoc_const(ExperimentLike, mini, "ASSAY_ORIENTATION")

library(S7)
library(s7contract)

# Shared S7 generics
area <- new_generic("area", "x")
draw <- new_generic("draw", "x")
perimeter <- new_generic("perimeter", "x")

# Domain classes
Circle <- new_class("Circle", properties = list(r = class_double))
Rect <- new_class("Rect", properties = list(w = class_double, h = class_double))

# Some methods
method(area, Circle) <- function(x) pi * x@r^2
method(draw, Circle) <- function(x) sprintf("circle(r = %s)", x@r)
method(area, Rect) <- function(x) x@w * x@h

# Go-like structural interface
Drawable <- new_interface("Drawable", generics = list(draw = draw))
Shape <- new_interface(
  "Shape",
  generics = list(area = area),
  parents = Drawable
)

stopifnot(implements(Circle, Shape))
stopifnot(!implements(Rect, Shape))
print(missing_requirements(Rect, Shape))

# Rust-like explicit traits
Measurable <- new_trait(
  "Measurable",
  methods = list(
    area = trait_method(area),
    perimeter = trait_method(perimeter, default = function(x) NA_real_)
  ),
  assoc_consts = c("UNITS")
)

impl_trait(
  Measurable,
  Circle,
  methods = list(area = function(x) pi * x@r^2),
  assoc_consts = list(UNITS = "unitless"),
  replace = TRUE
)

stopifnot(has_trait(Circle, Measurable))
stopifnot(!has_trait(Rect, Measurable))
stopifnot(is.na(trait_call(Measurable, "perimeter", Circle(r = 2))))
print(trait_assoc_const(Measurable, Circle, "UNITS"))

# Internal S7 classes for contract descriptors.
#
# These are intentionally not exported. User-facing construction should go
# through new_interface(), interface_requirement(), new_trait(), trait_method(),
# and impl_trait() so validation and method registration happen in one place.

s7_interface_requirement <- S7::new_class(
  "s7_interface_requirement",
  package = "s7contract",
  properties = list(
    name = S7::class_character,
    generic = S7::class_function,
    args = S7::class_list,
    returns = S7::class_any
  )
)

s7_interface <- S7::new_class(
  "s7_interface",
  package = "s7contract",
  properties = list(
    name = S7::class_character,
    package = S7::new_union(NULL, S7::class_character),
    parents = S7::class_list,
    requirements = S7::class_list
  )
)

s7_trait_method <- S7::new_class(
  "s7_trait_method",
  package = "s7contract",
  properties = list(
    name = S7::class_character,
    generic = S7::class_function,
    default = S7::new_union(NULL, S7::class_function),
    args = S7::class_list,
    returns = S7::class_any
  )
)

s7_assoc_item <- S7::new_class(
  "s7_assoc_item",
  package = "s7contract",
  properties = list(
    required = S7::class_logical,
    default = S7::class_any
  )
)

s7_trait <- S7::new_class(
  "s7_trait",
  package = "s7contract",
  properties = list(
    id = S7::class_character,
    name = S7::class_character,
    package = S7::new_union(NULL, S7::class_character),
    parents = S7::class_list,
    methods = S7::class_list,
    assoc_types = S7::class_list,
    assoc_consts = S7::class_list
  )
)

s7_trait_impl <- S7::new_class(
  "s7_trait_impl",
  package = "s7contract",
  properties = list(
    trait = s7_trait,
    trait_id = S7::class_character,
    trait_label = S7::class_character,
    target_class = S7::class_any,
    methods = S7::class_list,
    assoc_types = S7::class_list,
    assoc_consts = S7::class_list
  )
)

.is_interface_requirement <- function(x) {
  S7::S7_inherits(x, s7_interface_requirement)
}

.is_interface <- function(x) {
  S7::S7_inherits(x, s7_interface)
}

.is_trait_method <- function(x) {
  S7::S7_inherits(x, s7_trait_method)
}

.is_trait <- function(x) {
  S7::S7_inherits(x, s7_trait)
}

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup--------------------------------------------------------------------
library(S7)
library(s7contract)

## ----structural-interface-----------------------------------------------------
area <- new_generic("area", "x")
draw <- new_generic("draw", "x")

Circle <- new_class("Circle", properties = list(r = class_double))
Rect <- new_class("Rect", properties = list(w = class_double, h = class_double))

method(area, Circle) <- function(x) pi * x@r^2
method(draw, Circle) <- function(x) sprintf("circle(r = %s)", x@r)
method(area, Rect) <- function(x) x@w * x@h

Drawable <- new_interface("Drawable", generics = list(draw = draw))
Shape <- new_interface("Shape", generics = list(area = area), parents = Drawable)

implements(Circle, Shape)
implements(Rect, Shape)
missing_requirements(Rect, Shape)

## ----structural-consumer------------------------------------------------------
render <- function(x) {
  assert_implements(x, Drawable)
  draw(x)
}

render(Circle(r = 2))

## ----structural-mock----------------------------------------------------------
MockDrawable <- new_class("MockDrawable")
method(draw, MockDrawable) <- function(x) "mock drawing"

render(MockDrawable())

## ----dbi-like-class-family----------------------------------------------------
DatabaseConnection <- new_class("DatabaseConnection", abstract = TRUE)
MemoryConnection <- new_class(
  "MemoryConnection",
  parent = DatabaseConnection,
  properties = list(tables = class_list)
)

db_tables <- new_generic("db_tables", "con")
db_read_table <- new_generic(
  "db_read_table",
  "con",
  function(con, name) S7_dispatch()
)

method(db_tables, MemoryConnection) <- function(con) names(con@tables)
method(db_read_table, MemoryConnection) <- function(con, name) con@tables[[name]]

TableReader <- new_interface(
  "TableReader",
  generics = list(
    db_tables = interface_requirement(db_tables, returns = class_character),
    db_read_table = interface_requirement(
      db_read_table,
      args = list(name = class_character),
      returns = class_data.frame
    )
  )
)

first_table <- function(con) {
  assert_implements(con, TableReader)
  db_read_table(con, db_tables(con)[[1]])
}

con <- MemoryConnection(tables = list(iris = head(iris, 2)))
first_table(con)

## ----progressive-interface----------------------------------------------------
Canvas <- new_class("Canvas")

draw_on <- new_generic(
  "draw_on",
  c("x", "canvas"),
  function(x, canvas, position, ...) S7_dispatch()
)

method(draw_on, list(Circle, Canvas)) <- function(x, canvas, position, ...) {
  sprintf("circle(r = %s) at %s", x@r, position)
}

DrawableOnCanvas <- new_interface(
  "DrawableOnCanvas",
  generics = list(
    draw_on = interface_requirement(
      draw_on,
      args = list(canvas = Canvas, position = class_integer),
      returns = class_character
    )
  )
)

canvas <- Canvas()
circle <- Circle(r = 2)

implements(Circle, DrawableOnCanvas)
with(DrawableOnCanvas, draw_on(circle, canvas, position = 1L))
draw_on(circle, canvas, position = 1L) %::% DrawableOnCanvas

checked_draw <- with(DrawableOnCanvas, {
  function(x) draw_on(x, canvas, position = 1L)
})
checked_draw(circle)

## ----progressive-return-failure-----------------------------------------------
BadCircle <- new_class("BadCircle", properties = list(r = class_double))
method(draw_on, list(BadCircle, Canvas)) <- function(x, canvas, position, ...) {
  x@r
}

tryCatch(
  with(DrawableOnCanvas, draw_on(BadCircle(r = 2), canvas, position = 1L)),
  error = function(e) conditionMessage(e)
)

tryCatch(
  checked_draw(BadCircle(r = 2)),
  error = function(e) conditionMessage(e)
)

## ----number-interface---------------------------------------------------------
num_zero <- new_generic("num_zero", "x")
num_add <- new_generic("num_add", "x")
num_scale <- new_generic("num_scale", "x")

NumberLike <- new_interface(
  "NumberLike",
  generics = list(
    zero = num_zero,
    add = num_add,
    scale = num_scale
  )
)

method(num_zero, class_double) <- function(x) 0
method(num_add, class_double) <- function(x, y) x + y
method(num_scale, class_double) <- function(x, k) x * k

implements(class_double, NumberLike)
num_add(10, 5)
num_scale(10, 0.5)

## ----vector-interface---------------------------------------------------------
vec_length <- new_generic("vec_length", "x")
vec_slice <- new_generic("vec_slice", "x")
vec_values <- new_generic("vec_values", "x")

VectorLike <- new_interface(
  "VectorLike",
  generics = list(
    length = vec_length,
    slice = vec_slice,
    values = vec_values
  )
)

ReadDepth <- new_class(
  "ReadDepth",
  properties = list(
    position = class_integer,
    depth = class_double
  ),
  validator = function(self) {
    if (length(self@position) != length(self@depth)) {
      "@position and @depth must have the same length"
    }
  }
)

method(vec_length, ReadDepth) <- function(x) length(x@depth)
method(vec_slice, ReadDepth) <- function(x, i) {
  ReadDepth(position = x@position[i], depth = x@depth[i])
}
method(vec_values, ReadDepth) <- function(x) x@depth

coverage <- ReadDepth(
  position = 1:5,
  depth = c(12, 15, 9, 20, 17)
)

implements(coverage, VectorLike)
vec_values(vec_slice(coverage, 2:4))

## ----vector-consumer----------------------------------------------------------
window_mean <- function(x, i) {
  assert_implements(x, VectorLike)
  mean(vec_values(vec_slice(x, i)))
}

window_mean(coverage, 2:4)

## ----trait--------------------------------------------------------------------
perimeter <- new_generic("perimeter", "x")

Measurable <- new_trait(
  "Measurable",
  methods = list(
    area = trait_method(area),
    perimeter = trait_method(perimeter, default = function(x) NA_real_)
  ),
  assoc_consts = c("UNITS")
)

impl_trait(
  Measurable,
  Circle,
  methods = list(area = function(x) pi * x@r^2),
  assoc_consts = list(UNITS = "unitless"),
  replace = TRUE
)

has_trait(Circle, Measurable)
trait_call(Measurable, "area", Circle(r = 2))
trait_call(Measurable, "perimeter", Circle(r = 2))
trait_assoc_const(Measurable, Circle, "UNITS")

## ----monad-dictionary---------------------------------------------------------
Maybe <- new_class("Maybe", abstract = TRUE)
Nothing <- new_class("Nothing", parent = Maybe)
Just <- new_class("Just", parent = Maybe, properties = list(value = class_any))

MonadDict <- new_class(
  "MonadDict",
  properties = list(
    name = class_character,
    pure = class_function,
    bind = class_function
  )
)

dict_pure <- new_generic("dict_pure", "x")
dict_bind <- new_generic("dict_bind", "x")

MonadDictionary <- new_interface(
  "MonadDictionary",
  generics = list(
    pure = dict_pure,
    bind = dict_bind
  )
)

method(dict_pure, MonadDict) <- function(x, value) {
  (x@pure)(value)
}
method(dict_bind, MonadDict) <- function(x, mx, f) {
  (x@bind)(mx, f)
}

MaybeMonad <- MonadDict(
  name = "Maybe",
  pure = function(value) Just(value = value),
  bind = function(mx, f) {
    if (S7_inherits(mx, Nothing)) {
      Nothing()
    } else {
      f(mx@value)
    }
  }
)

implements(MaybeMonad, MonadDictionary)

## ----monad-dictionary-use-----------------------------------------------------
dict_bind(
  MaybeMonad,
  Just(value = 2),
  function(x) dict_pure(MaybeMonad, x + 1)
)

dict_bind(
  MaybeMonad,
  Nothing(),
  function(x) dict_pure(MaybeMonad, x + 1)
)

## ----monad-laws---------------------------------------------------------------
maybe_equal <- function(x, y) {
  if (S7_inherits(x, Nothing) && S7_inherits(y, Nothing)) {
    return(TRUE)
  }
  if (S7_inherits(x, Just) && S7_inherits(y, Just)) {
    return(identical(x@value, y@value))
  }
  FALSE
}

f <- function(x) dict_pure(MaybeMonad, x + 1)
g <- function(x) dict_pure(MaybeMonad, x * 2)
mx <- Just(value = 10)

c(
  left_identity = maybe_equal(
    dict_bind(MaybeMonad, dict_pure(MaybeMonad, 10), f),
    f(10)
  ),
  right_identity = maybe_equal(
    dict_bind(MaybeMonad, mx, function(x) dict_pure(MaybeMonad, x)),
    mx
  ),
  associativity = maybe_equal(
    dict_bind(MaybeMonad, dict_bind(MaybeMonad, mx, f), g),
    dict_bind(MaybeMonad, mx, function(x) dict_bind(MaybeMonad, f(x), g))
  )
)

.s7contract_registry <- new.env(parent = emptyenv())
.s7contract_registry$next_trait_id <- 0L
.s7contract_registry$impls <- list()

.next_trait_id <- function() {
  .s7contract_registry$next_trait_id <- .s7contract_registry$next_trait_id + 1L
  sprintf("trait:%d", .s7contract_registry$next_trait_id)
}

#' Build a Rust-like explicit trait on top of S7
#'
#' `new_trait()` adds a nominal contract registry on top of S7 dispatch. A class
#' only has the trait after `impl_trait()` records the implementation, even if
#' compatible S7 methods already exist.
#'
#' This makes default methods and associated metadata practical, but the result
#' remains a runtime R abstraction. It does not emulate Rust's compile-time
#' trait bounds, coherence, orphan rules, or type-checked associated types.
#'
#' @param name For `new_trait()`, the trait name. For `trait_method()`, the
#'   method name; it defaults to the generic name when omitted.
#' @param methods For `new_trait()`, a named list of S7 generics or
#'   `trait_method()` objects.
#' @param parents Optional trait or list of supertraits.
#' @param assoc_types Required associated type names, or a named list of default
#'   associated type values.
#' @param assoc_consts Required associated constant names, or a named list of
#'   default constant values.
#' @param package Optional package name used only for display.
#' @return `new_trait()` returns an S7 object of class `s7_trait`.
#'   `trait_method()` returns an S7 object of class `s7_trait_method`.
#' @examples
#' local({
#'   area <- S7::new_generic("area", "x")
#'   perimeter <- S7::new_generic("perimeter", "x")
#'
#'   Circle <- S7::new_class(
#'     "Circle",
#'     properties = list(r = S7::class_double)
#'   )
#'
#'   Measurable <- new_trait(
#'     "Measurable",
#'     methods = list(
#'       area = trait_method(area),
#'       perimeter = trait_method(perimeter, default = function(x) NA_real_)
#'     ),
#'     assoc_consts = c("UNITS")
#'   )
#'
#'   impl_trait(
#'     Measurable,
#'     Circle,
#'     methods = list(area = function(x) pi * x@r^2),
#'     assoc_consts = list(UNITS = "unitless")
#'   )
#'
#'   has_trait(Circle, Measurable)
#'   trait_call(Measurable, "area", Circle(r = 2))
#'   trait_assoc_const(Measurable, Circle, "UNITS")
#' })
#' @export
new_trait <- function(
  name,
  methods = list(),
  parents = list(),
  assoc_types = character(),
  assoc_consts = list(),
  package = NULL
) {
  if (!is.character(name) || length(name) != 1 || !nzchar(name)) {
    .abort("`name` must be a non-empty string.")
  }
  if (!is.null(package) && (!is.character(package) || length(package) != 1)) {
    .abort("`package` must be NULL or a single string.")
  }

  s7_trait(
    id = .next_trait_id(),
    name = name,
    package = package,
    parents = .normalise_trait_parents(parents),
    methods = .normalise_trait_methods(methods),
    assoc_types = .normalise_assoc(assoc_types, "assoc_types"),
    assoc_consts = .normalise_assoc(assoc_consts, "assoc_consts")
  )
}

#' @param generic An S7 generic function.
#' @param default Optional default implementation. If supplied, `impl_trait()`
#'   uses it when a class does not provide an override for that method.
#' @param args Optional named list of S7 classes, interfaces, or traits for
#'   runtime argument checking with `with()` or `%::%`. Dispatch arguments other
#'   than the first can use S7 classes or unions to refine multiple-dispatch
#'   requirements.
#' @param returns Optional S7 class, interface, or trait for runtime return
#'   checking with `with()` or `%::%`; defaults to `S7::class_any`.
#' @rdname new_trait
#' @export
trait_method <- function(
  generic,
  default = NULL,
  name = NULL,
  args = list(),
  returns = S7::class_any
) {
  .check_s7_generic(generic, "generic")
  if (!is.null(default) && !is.function(default)) {
    .abort("`default` must be NULL or a function.")
  }
  if (is.null(name)) {
    name <- .generic_label(generic)
  }
  if (!is.character(name) || length(name) != 1 || !nzchar(name)) {
    .abort("`name` must be a non-empty string.")
  }

  s7_trait_method(
    name = name,
    generic = generic,
    default = default,
    args = .normalise_type_specs(args, "args"),
    returns = .normalise_return_spec(returns)
  )
}

.as_trait_method <- function(x, name = NULL) {
  if (.is_trait_method(x)) {
    if (!is.null(name)) {
      x@name <- name
    }
    return(x)
  }
  if (is.function(x)) {
    return(trait_method(x, name = name))
  }
  .abort("Trait methods must be S7 generics or trait_method() objects.")
}

.normalise_trait_methods <- function(methods) {
  if (is.null(methods)) {
    methods <- list()
  }
  if (is.function(methods) || .is_trait_method(methods)) {
    methods <- list(methods)
  }
  if (!is.list(methods)) {
    .abort("`methods` must be a list of S7 generics or trait_method() objects.")
  }

  nms <- names(methods)
  if (is.null(nms)) {
    nms <- rep("", length(methods))
  }

  out <- vector("list", length(methods))
  for (i in seq_along(methods)) {
    nm <- if (nzchar(nms[[i]])) nms[[i]] else NULL
    method <- .as_trait_method(methods[[i]], name = nm)
    out[[i]] <- method
    nms[[i]] <- method@name
  }
  names(out) <- nms
  out
}

.normalise_trait_parents <- function(parents) {
  if (is.null(parents)) {
    return(list())
  }
  if (.is_trait(parents)) {
    parents <- list(parents)
  }
  if (!is.list(parents)) {
    .abort("`parents` must be a trait or a list of traits.")
  }
  for (parent in parents) {
    if (!.is_trait(parent)) {
      .abort("Every parent must be created with new_trait().")
    }
  }
  parents
}

.normalise_assoc <- function(x, what) {
  if (is.null(x)) {
    return(list())
  }
  if (is.character(x)) {
    out <- lapply(x, function(name) {
      s7_assoc_item(required = TRUE, default = NULL)
    })
    names(out) <- x
    return(out)
  }
  if (is.list(x)) {
    if (length(x) > 0 && (is.null(names(x)) || any(names(x) == ""))) {
      .abort("`%s` must be a named list or a character vector.", what)
    }
    out <- lapply(x, function(value) {
      s7_assoc_item(required = FALSE, default = value)
    })
    return(out)
  }
  .abort("`%s` must be NULL, a character vector, or a named list.", what)
}

.trait_label <- function(trait) {
  if (!is.null(trait@package)) {
    sprintf("%s::%s", trait@package, trait@name)
  } else {
    trait@name
  }
}

#' Inspect or use a Rust-like explicit trait
#'
#' @param trait A trait created by `new_trait()`.
#' @param inherited Include inherited methods from supertraits?
#' @return `trait_methods()` returns a named list of `trait_method()` objects.
#'   `impl_trait()` returns the stored implementation record, invisibly.
#'   `trait_report()` returns a one-row data frame. `has_trait()` returns a
#'   single logical value. `assert_trait()` returns `x`, unchanged.
#'   `trait_call()` returns the result of the underlying S7 generic.
#'   `trait_assoc_type()` and `trait_assoc_const()` return the stored associated
#'   item value.
#' @rdname trait_methods
#' @export
trait_methods <- function(trait, inherited = TRUE) {
  if (!.is_trait(trait)) {
    .abort("`trait` must be created with new_trait().")
  }

  out <- list()
  if (isTRUE(inherited)) {
    for (parent in trait@parents) {
      out <- c(out, trait_methods(parent, inherited = TRUE))
    }
  }
  out <- c(out, trait@methods)

  if (length(out) > 0) {
    out <- out[!duplicated(names(out), fromLast = TRUE)]
  }
  out
}

.trait_assoc_types <- function(trait, inherited = TRUE) {
  out <- list()
  if (isTRUE(inherited)) {
    for (parent in trait@parents) {
      out <- c(out, .trait_assoc_types(parent, inherited = TRUE))
    }
  }
  out <- c(out, trait@assoc_types)
  if (length(out) > 0) out[!duplicated(names(out), fromLast = TRUE)] else out
}

.trait_assoc_consts <- function(trait, inherited = TRUE) {
  out <- list()
  if (isTRUE(inherited)) {
    for (parent in trait@parents) {
      out <- c(out, .trait_assoc_consts(parent, inherited = TRUE))
    }
  }
  out <- c(out, trait@assoc_consts)
  if (length(out) > 0) out[!duplicated(names(out), fromLast = TRUE)] else out
}

.find_trait_impl <- function(trait, class) {
  impls <- .s7contract_registry$impls
  for (impl in impls) {
    if (
      identical(impl@trait_id, trait@id) &&
        .class_equal(impl@target_class, class)
    ) {
      return(impl)
    }
  }
  NULL
}

.store_trait_impl <- function(impl, replace = FALSE) {
  impls <- .s7contract_registry$impls
  keep <- rep(TRUE, length(impls))

  for (i in seq_along(impls)) {
    if (
      identical(impls[[i]]@trait_id, impl@trait_id) &&
        .class_equal(impls[[i]]@target_class, impl@target_class)
    ) {
      if (!replace) {
        .abort(
          "%s is already implemented for %s. Pass replace = TRUE to replace it.",
          impl@trait_label,
          .class_label(impl@target_class)
        )
      }
      keep[[i]] <- FALSE
    }
  }

  .s7contract_registry$impls <- c(impls[keep], list(impl))
  invisible(impl)
}

.normalise_impl_methods <- function(methods) {
  if (is.null(methods)) {
    return(list())
  }
  if (!is.list(methods)) {
    .abort("`methods` must be a named list of functions.")
  }
  if (
    length(methods) > 0 &&
      (is.null(names(methods)) || any(names(methods) == ""))
  ) {
    .abort("`methods` must be a named list of functions.")
  }
  for (name in names(methods)) {
    if (!is.function(methods[[name]])) {
      .abort("Implementation for `%s` must be a function.", name)
    }
  }
  methods
}

.resolve_assoc_impl <- function(required, provided, what) {
  if (is.null(provided)) {
    provided <- list()
  }
  if (!is.list(provided)) {
    .abort("`%s` must be a named list.", what)
  }
  if (
    length(provided) > 0 &&
      (is.null(names(provided)) || any(names(provided) == ""))
  ) {
    .abort("`%s` must be a named list.", what)
  }

  out <- list()
  for (name in names(required)) {
    spec <- required[[name]]
    if (name %in% names(provided)) {
      out[name] <- list(provided[[name]])
    } else if (isTRUE(spec@required)) {
      .abort("Missing required associated item `%s` in `%s`.", name, what)
    } else {
      out[name] <- list(spec@default)
    }
  }

  extra <- setdiff(names(provided), names(required))
  if (length(extra) > 0) {
    .abort(
      "Unknown associated item(s) for `%s`: %s",
      what,
      paste(extra, collapse = ", ")
    )
  }
  out
}

#' @param class An S7 class or base class wrapper.
#' @param methods Named list of method implementations. Omitted trait methods
#'   use their default implementation when one is available.
#' @param assoc_types Named list of associated type values.
#' @param assoc_consts Named list of associated constant values.
#' @param replace Replace an existing implementation record and silence warnings
#'   about visible S7 methods?
#' @rdname trait_methods
#' @export
impl_trait <- function(
  trait,
  class,
  methods = list(),
  assoc_types = list(),
  assoc_consts = list(),
  replace = FALSE
) {
  if (!.is_trait(trait)) {
    .abort("`trait` must be created with new_trait().")
  }

  cls <- .as_class_or_null(class, arg = "class")
  if (is.null(cls)) {
    .abort(
      "`class` must be an S7 class, S7 union, S3 class wrapper, S4 class, or base class wrapper."
    )
  }

  for (parent in trait@parents) {
    if (is.null(.find_trait_impl(parent, cls))) {
      .abort(
        "Cannot implement %s for %s until its supertrait %s is implemented.",
        .trait_label(trait),
        .class_label(cls),
        .trait_label(parent)
      )
    }
  }

  if (!replace && !is.null(.find_trait_impl(trait, cls))) {
    .abort(
      "%s is already implemented for %s. Pass replace = TRUE to replace it.",
      .trait_label(trait),
      .class_label(cls)
    )
  }

  trait_reqs <- trait_methods(trait, inherited = FALSE)
  provided_methods <- .normalise_impl_methods(methods)
  resolved_methods <- list()

  for (name in names(trait_reqs)) {
    req <- trait_reqs[[name]]
    fun <- provided_methods[[name]]
    if (is.null(fun)) {
      fun <- req@default
    }
    if (is.null(fun)) {
      .abort(
        "Missing required trait method `%s` for %s.",
        name,
        .trait_label(trait)
      )
    }
    resolved_methods[[name]] <- fun
  }

  extra_methods <- setdiff(names(provided_methods), names(trait_reqs))
  if (length(extra_methods) > 0) {
    .abort("Unknown trait method(s): %s", paste(extra_methods, collapse = ", "))
  }

  resolved_assoc_types <- .resolve_assoc_impl(
    .trait_assoc_types(trait, inherited = FALSE),
    assoc_types,
    "assoc_types"
  )
  resolved_assoc_consts <- .resolve_assoc_impl(
    .trait_assoc_consts(trait, inherited = FALSE),
    assoc_consts,
    "assoc_consts"
  )

  impl <- s7_trait_impl(
    trait = trait,
    trait_id = trait@id,
    trait_label = .trait_label(trait),
    target_class = cls,
    methods = resolved_methods,
    assoc_types = resolved_assoc_types,
    assoc_consts = resolved_assoc_consts
  )

  for (name in names(trait_reqs)) {
    .register_s7_method(
      generic = trait_reqs[[name]]@generic,
      class = .requirement_signature(trait_reqs[[name]], cls),
      fun = resolved_methods[[name]],
      replace = replace
    )
  }

  .store_trait_impl(impl, replace = replace)

  invisible(impl)
}

#' @param x An object or class.
#' @rdname trait_methods
#' @export
trait_report <- function(x, trait) {
  if (!.is_trait(trait)) {
    .abort("`trait` must be created with new_trait().")
  }

  cls <- .target_class_or_null(x, arg = "x")
  if (is.null(cls)) {
    return(data.frame(
      trait = .trait_label(trait),
      class = "<unknown>",
      ok = FALSE,
      message = "Could not determine an S7 class for this value.",
      stringsAsFactors = FALSE
    ))
  }

  impl <- .find_trait_impl(trait, cls)
  data.frame(
    trait = .trait_label(trait),
    class = .class_label(cls),
    ok = !is.null(impl),
    message = if (is.null(impl)) "No explicit impl_trait() record." else "",
    stringsAsFactors = FALSE
  )
}

#' @rdname trait_methods
#' @export
has_trait <- function(x, trait) {
  trait_report(x, trait)$ok[[1]]
}

#' @param arg Name to use in error messages.
#' @rdname trait_methods
#' @export
assert_trait <- function(x, trait, arg = deparse(substitute(x))) {
  report <- trait_report(x, trait)
  if (!report$ok[[1]]) {
    .abort(
      "%s does not explicitly implement %s for %s: %s",
      arg,
      report$trait[[1]],
      report$class[[1]],
      report$message[[1]]
    )
  }
  x
}

#' @param method Method name within the trait.
#' @param ... Additional arguments passed to the S7 generic.
#' @rdname trait_methods
#' @export
trait_call <- function(trait, method, x, ...) {
  if (!.is_trait(trait)) {
    .abort("`trait` must be created with new_trait().")
  }
  if (!is.character(method) || length(method) != 1 || !nzchar(method)) {
    .abort("`method` must be a non-empty string.")
  }

  assert_trait(x, trait)
  reqs <- trait_methods(trait, inherited = TRUE)
  if (!method %in% names(reqs)) {
    .abort("Trait %s has no method `%s`.", .trait_label(trait), method)
  }
  reqs[[method]]@generic(x, ...)
}

.assoc_value_from_impl <- function(trait, cls, field, name, impl = NULL) {
  if (is.null(impl)) {
    impl <- .find_trait_impl(trait, cls)
  }

  if (!is.null(impl)) {
    values <- S7::prop(impl, field)
    if (name %in% names(values)) {
      return(list(ok = TRUE, value = values[[name]]))
    }
  }

  for (parent in trait@parents) {
    found <- .assoc_value_from_impl(parent, cls, field, name)
    if (isTRUE(found$ok)) {
      return(found)
    }
  }

  list(ok = FALSE, value = NULL)
}

.assoc_from_impl <- function(trait, x, field, name) {
  cls <- .target_class_or_null(x, arg = "x")
  if (is.null(cls)) {
    .abort("Could not determine an S7 class for this value.")
  }
  impl <- .find_trait_impl(trait, cls)
  if (is.null(impl)) {
    .abort(
      "%s does not explicitly implement %s.",
      .class_label(cls),
      .trait_label(trait)
    )
  }

  found <- .assoc_value_from_impl(trait, cls, field, name, impl = impl)
  if (!isTRUE(found$ok)) {
    .abort("Trait %s has no associated item `%s`.", .trait_label(trait), name)
  }
  found$value
}

#' @param name Associated item name.
#' @rdname trait_methods
#' @export
trait_assoc_type <- function(trait, x, name) {
  .assoc_from_impl(trait, x, "assoc_types", name)
}

#' @rdname trait_methods
#' @export
trait_assoc_const <- function(trait, x, name) {
  .assoc_from_impl(trait, x, "assoc_consts", name)
}

.print_s7_trait <- function(x, ...) {
  reqs <- trait_methods(x, inherited = TRUE)
  assoc_types <- .trait_assoc_types(x, inherited = TRUE)
  assoc_consts <- .trait_assoc_consts(x, inherited = TRUE)

  cat(sprintf("<S7 Rust-like trait> %s\n", .trait_label(x)))
  if (length(x@parents) > 0) {
    cat(
      "  supertraits:",
      paste(vapply(x@parents, .trait_label, character(1)), collapse = ", "),
      "\n"
    )
  }
  cat("  methods:")
  if (length(reqs) == 0) {
    cat(" <none>\n")
  } else {
    cat("\n")
    for (req in reqs) {
      suffix <- if (is.null(req@default)) "" else " [default]"
      typed_args <- names(req@args)
      typed_args <- if (length(typed_args) == 0) {
        ""
      } else {
        sprintf(" args: %s", paste(typed_args, collapse = ", "))
      }
      cat(sprintf("    - %s()%s%s\n", req@name, suffix, typed_args))
    }
  }
  if (length(assoc_types) > 0) {
    cat("  associated types:", paste(names(assoc_types), collapse = ", "), "\n")
  }
  if (length(assoc_consts) > 0) {
    cat(
      "  associated consts:",
      paste(names(assoc_consts), collapse = ", "),
      "\n"
    )
  }
  invisible(x)
}

library(S7)

local({
  area <- new_generic("area_trait_test", "x")
  perimeter <- new_generic("perimeter_trait_test", "x")
  draw <- new_generic("draw_trait_test", "x")

  Circle <- new_class(
    "CircleTraitTest",
    properties = list(r = class_double)
  )

  Measurable <- new_trait(
    "MeasurableTraitTest",
    methods = list(
      area = trait_method(area),
      perimeter = trait_method(perimeter, default = function(x) NA_real_)
    ),
    assoc_consts = c("UNITS")
  )

  expect_equal(names(trait_methods(Measurable)), c("area", "perimeter"))
  expect_true(any(grepl(
    "<S7 Rust-like trait> MeasurableTraitTest",
    capture.output(print(Measurable)),
    fixed = TRUE
  )))
  expect_false(has_trait(Circle, Measurable))

  impl <- impl_trait(
    Measurable,
    Circle,
    methods = list(area = function(x) pi * x@r^2),
    assoc_consts = list(UNITS = "unitless"),
    replace = TRUE
  )

  expect_true(has_trait(Circle, Measurable))
  expect_equal(trait_call(Measurable, "area", Circle(r = 2)), pi * 4)
  expect_true(is.na(trait_call(Measurable, "perimeter", Circle(r = 2))))
  expect_equal(trait_assoc_const(Measurable, Circle, "UNITS"), "unitless")

  Drawable <- new_trait(
    "DrawableTraitTest",
    methods = list(draw = trait_method(draw))
  )
  Renderable <- new_trait(
    "RenderableTraitTest",
    parents = list(Drawable)
  )

  expect_error(
    impl_trait(Renderable, Circle, replace = TRUE),
    "supertrait"
  )

  impl_trait(
    Drawable,
    Circle,
    methods = list(draw = function(x) sprintf("circle(%s)", x@r)),
    replace = TRUE
  )
  impl_trait(Renderable, Circle, replace = TRUE)
  expect_true(has_trait(Circle, Renderable))

  bad <- new_generic("bad_trait_test", "x")
  BadTrait <- new_trait(
    "BadTraitTest",
    methods = list(bad = trait_method(bad))
  )

  expect_error(
    impl_trait(BadTrait, Circle, methods = list(bad = function(y) y)),
    "dispatches on"
  )
  expect_false(has_trait(Circle, BadTrait))
  impl_trait(BadTrait, Circle, methods = list(bad = function(x) "ok"))
  expect_true(has_trait(Circle, BadTrait))
  expect_equal(trait_call(BadTrait, "bad", Circle(r = 1)), "ok")

  ParentAssoc <- new_trait(
    "ParentAssocTraitTest",
    assoc_consts = c("PARENT")
  )
  ChildAssoc <- new_trait(
    "ChildAssocTraitTest",
    parents = list(ParentAssoc),
    assoc_consts = list(LOCAL_NULL = NULL)
  )
  RequiredNull <- new_trait(
    "RequiredNullTraitTest",
    assoc_consts = c("VALUE")
  )

  impl_trait(
    ParentAssoc,
    Circle,
    assoc_consts = list(PARENT = "parent"),
    replace = TRUE
  )
  impl_trait(ChildAssoc, Circle, replace = TRUE)
  impl_trait(
    RequiredNull,
    Circle,
    assoc_consts = list(VALUE = NULL),
    replace = TRUE
  )

  expect_equal(trait_assoc_const(ChildAssoc, Circle, "PARENT"), "parent")
  expect_null(trait_assoc_const(ChildAssoc, Circle, "LOCAL_NULL"))
  expect_null(trait_assoc_const(RequiredNull, Circle, "VALUE"))
})

# Small internal helpers. Kept dependency-free on purpose.

.abort <- function(..., call. = FALSE) {
  stop(sprintf(...), call. = call.)
}

.warn <- function(..., call. = FALSE) {
  warning(sprintf(...), call. = call.)
}

.as_class_or_null <- function(x, arg = "x") {
  tryCatch(S7::as_class(x, arg = arg), error = function(e) NULL)
}

.is_s7_generic <- function(x) {
  is.function(x) && inherits(x, "S7_generic")
}

.check_s7_generic <- function(x, arg = "generic") {
  if (!.is_s7_generic(x)) {
    .abort("`%s` must be an S7 generic created with S7::new_generic().", arg)
  }
  invisible(x)
}

.class_key <- function(class) {
  cls <- .as_class_or_null(class, arg = "class")
  if (is.null(cls)) {
    return(NA_character_)
  }

  nm <- tryCatch(nameOfClass(cls), error = function(e) NULL)
  if (!is.null(nm) && length(nm) == 1 && !is.na(nm) && nzchar(nm)) {
    return(nm)
  }

  line <- paste(utils::capture.output(print(cls)), collapse = " ")
  line <- gsub("\\s+", " ", line)
  trimws(line)
}

.class_equal <- function(a, b) {
  identical(a, b) || identical(.class_key(a), .class_key(b))
}

.class_label <- function(class) {
  key <- .class_key(class)
  if (is.na(key) || key == "") "<unknown class>" else sprintf("<%s>", key)
}

.base_class_of <- function(x) {
  switch(
    typeof(x),
    logical = S7::class_logical,
    integer = S7::class_integer,
    double = S7::class_double,
    complex = S7::class_complex,
    character = S7::class_character,
    raw = S7::class_raw,
    list = S7::class_list,
    expression = S7::class_expression,
    symbol = S7::class_name,
    language = S7::class_call,
    closure = S7::class_function,
    NULL
  )
}

.target_class_or_null <- function(x, arg = "x") {
  cls <- .as_class_or_null(x, arg = arg)
  if (!is.null(cls)) {
    return(cls)
  }

  cls <- tryCatch(S7::S7_class(x), error = function(e) NULL)
  if (!is.null(cls)) {
    return(cls)
  }

  .base_class_of(x)
}

.generic_label <- function(generic, fallback = "method") {
  if (inherits(generic, "S7_generic")) {
    return(generic@name)
  }

  line <- paste(utils::capture.output(print(generic))[1], collapse = "")
  name <- sub("^.*<S7_generic>\\s*", "", line)
  name <- sub("\\(.*$", "", name)
  name <- trimws(name)
  if (nzchar(name) && !identical(name, line)) {
    return(name)
  }

  fallback
}

.register_s7_method <- function(generic, class, fun, replace = FALSE) {
  if (!is.function(fun)) {
    .abort("S7 method implementation must be a function")
  }

  if (!replace) {
    existing <- tryCatch(
      S7::method(generic, class = class),
      error = function(e) NULL
    )
    if (!is.null(existing)) {
      .warn(
        "An S7 method for %s and %s is already visible; registering anyway. Pass replace = TRUE to silence this warning.",
        .generic_label(generic),
        .class_label(class)
      )
    }
  }

  S7::`method<-`(generic, class, value = fun)
  invisible(fun)
}

#' Build a Go-like structural interface on top of S7
#'
#' `new_interface()` models the method-list part of Go interfaces as a list of
#' required S7 generics. An interface is just a named set of required generics,
#' and a class or object satisfies it when S7 can find a method for every
#' required generic.
#'
#' This deliberately mirrors Go's basic interfaces defined only by methods. The
#' intended style is to define small interfaces at the point where consuming code
#' needs a behavior, not beside a single concrete implementation. Define S7
#' classes, generics, and methods normally; then let consumers name the protocol
#' they accept. Up-front interfaces can still be useful for deliberate package
#' protocols, abstract data types, or recursive protocols.
#'
#' It does not attempt to emulate Go's full post-1.18 type-set language such as
#' tilde type terms, unions of concrete types, or pointer/value receiver rules.
#'
#' @param name For `new_interface()`, the interface name. For
#'   `interface_requirement()`, the requirement name; it defaults to the generic
#'   name when omitted.
#' @param generics For `new_interface()`, a named list of S7 generics or
#'   `interface_requirement()` objects. These are generic functions because S7
#'   methods are registered separately on generics.
#' @param methods Compatibility alias for `generics`.
#' @param parents Optional interface or list of interfaces to embed.
#' @param package Optional package name used only for display.
#' @return `new_interface()` returns an S7 object of class
#'   `s7_interface`. `interface_requirement()` returns an S7 object of class
#'   `s7_interface_requirement`.
#' @examples
#' local({
#'   area <- S7::new_generic("area", "x")
#'   draw <- S7::new_generic("draw", "x")
#'
#'   Circle <- S7::new_class(
#'     "Circle",
#'     properties = list(r = S7::class_double)
#'   )
#'   Rect <- S7::new_class(
#'     "Rect",
#'     properties = list(w = S7::class_double, h = S7::class_double)
#'   )
#'
#'   S7::method(area, Circle) <- function(x) pi * x@r^2
#'   S7::method(draw, Circle) <- function(x) sprintf("circle(r = %s)", x@r)
#'   S7::method(area, Rect) <- function(x) x@w * x@h
#'
#'   Drawable <- new_interface("Drawable", generics = list(draw = draw))
#'   Shape <- new_interface("Shape", generics = list(area = area), parents = Drawable)
#'
#'   implements(Circle, Shape)
#'   missing_requirements(Rect, Shape)
#' })
#' @export
new_interface <- function(
  name,
  generics = list(),
  parents = list(),
  package = NULL,
  methods = NULL
) {
  if (!is.null(methods)) {
    if (!missing(generics)) {
      .abort("Use either `generics` or `methods`, not both.")
    }
    generics <- methods
  }

  if (!is.character(name) || length(name) != 1 || !nzchar(name)) {
    .abort("`name` must be a non-empty string.")
  }
  if (!is.null(package) && (!is.character(package) || length(package) != 1)) {
    .abort("`package` must be NULL or a single string.")
  }

  s7_interface(
    name = name,
    package = package,
    parents = .normalise_interface_parents(parents),
    requirements = .normalise_interface_generics(generics)
  )
}

#' @param generic An S7 generic function.
#' @param args Optional named list of S7 classes, interfaces, or traits for
#'   runtime argument checking with `with()` or `%::%`. Arguments named in
#'   `args` are also checked against generic and method formals during
#'   conformance checks. Dispatch arguments other than the first can use S7
#'   classes or unions to refine multiple-dispatch requirements.
#' @param returns Optional S7 class, interface, or trait for runtime return
#'   checking with `with()` or `%::%`; defaults to `S7::class_any`.
#' @rdname new_interface
#' @export
interface_requirement <- function(
  generic,
  name = NULL,
  args = list(),
  returns = S7::class_any
) {
  .check_s7_generic(generic, "generic")
  if (is.null(name)) {
    name <- .generic_label(generic)
  }
  if (!is.character(name) || length(name) != 1 || !nzchar(name)) {
    .abort("`name` must be a non-empty string.")
  }

  s7_interface_requirement(
    name = name,
    generic = generic,
    args = .normalise_type_specs(args, "args"),
    returns = .normalise_return_spec(returns)
  )
}

.as_interface_requirement <- function(x, name = NULL) {
  if (.is_interface_requirement(x)) {
    if (!is.null(name)) {
      x@name <- name
    }
    return(x)
  }
  if (is.function(x)) {
    return(interface_requirement(x, name = name))
  }
  .abort(
    "Interface requirements must be S7 generics or interface_requirement() objects."
  )
}

.normalise_interface_generics <- function(generics) {
  if (is.null(generics)) {
    generics <- list()
  }
  if (is.function(generics) || .is_interface_requirement(generics)) {
    generics <- list(generics)
  }
  if (!is.list(generics)) {
    .abort(
      "`generics` must be a list of S7 generics or interface_requirement() objects."
    )
  }

  nms <- names(generics)
  if (is.null(nms)) {
    nms <- rep("", length(generics))
  }

  out <- vector("list", length(generics))
  for (i in seq_along(generics)) {
    nm <- if (nzchar(nms[[i]])) nms[[i]] else NULL
    req <- .as_interface_requirement(generics[[i]], name = nm)
    out[[i]] <- req
    nms[[i]] <- req@name
  }
  names(out) <- nms
  out
}

.normalise_interface_parents <- function(parents) {
  if (is.null(parents)) {
    return(list())
  }
  if (.is_interface(parents)) {
    parents <- list(parents)
  }
  if (!is.list(parents)) {
    .abort("`parents` must be an interface or a list of interfaces.")
  }
  for (parent in parents) {
    if (!.is_interface(parent)) {
      .abort("Every parent must be created with new_interface().")
    }
  }
  parents
}

.interface_label <- function(interface) {
  if (!is.null(interface@package)) {
    sprintf("%s::%s", interface@package, interface@name)
  } else {
    interface@name
  }
}

#' Inspect or check a Go-like structural interface
#'
#' @param interface An interface created by `new_interface()`.
#' @param inherited Include inherited requirements from parent interfaces?
#' @return `interface_requirements()` returns a named list of
#'   `interface_requirement()` objects. `interface_report()` and
#'   `missing_requirements()` return data frames. `implements()` returns a single
#'   logical value. `assert_implements()` and `as_interface()` return `x`,
#'   unchanged.
#' @rdname interface_requirements
#' @export
interface_requirements <- function(interface, inherited = TRUE) {
  if (!.is_interface(interface)) {
    .abort("`interface` must be created with new_interface().")
  }

  out <- list()
  if (isTRUE(inherited)) {
    for (parent in interface@parents) {
      out <- c(out, interface_requirements(parent, inherited = TRUE))
    }
  }
  out <- c(out, interface@requirements)

  if (length(out) > 0) {
    out <- out[!duplicated(names(out), fromLast = TRUE)]
  }
  out
}

#' @param x An object, or an S7 class/base class wrapper.
#' @rdname interface_requirements
#' @export
interface_report <- function(x, interface) {
  reqs <- interface_requirements(interface, inherited = TRUE)

  rows <- lapply(reqs, function(req) {
    found <- .lookup_requirement_method(req, x)
    data.frame(
      interface = .interface_label(interface),
      requirement = req@name,
      ok = found$ok,
      message = if (found$ok) "" else conditionMessage(found$error),
      stringsAsFactors = FALSE
    )
  })

  if (length(rows) == 0) {
    return(data.frame(
      interface = character(),
      requirement = character(),
      ok = logical(),
      message = character(),
      stringsAsFactors = FALSE
    ))
  }

  do.call(rbind, rows)
}

#' @rdname interface_requirements
#' @export
missing_requirements <- function(x, interface) {
  report <- interface_report(x, interface)
  report[!report$ok, , drop = FALSE]
}

#' @rdname interface_requirements
#' @export
implements <- function(x, interface) {
  nrow(missing_requirements(x, interface)) == 0L
}

#' @param arg Name to use in error messages.
#' @rdname interface_requirements
#' @export
assert_implements <- function(x, interface, arg = deparse(substitute(x))) {
  miss <- missing_requirements(x, interface)
  if (nrow(miss) > 0) {
    .abort(
      "%s does not implement %s; missing: %s",
      arg,
      .interface_label(interface),
      paste0(miss$requirement, "()", collapse = ", ")
    )
  }
  x
}

#' @rdname interface_requirements
#' @export
as_interface <- function(x, interface) {
  assert_implements(x, interface)
}

.print_s7_interface <- function(x, ...) {
  reqs <- interface_requirements(x, inherited = TRUE)
  cat(sprintf("<S7 Go-like interface> %s\n", .interface_label(x)))
  if (length(x@parents) > 0) {
    cat(
      "  embeds:",
      paste(vapply(x@parents, .interface_label, character(1)), collapse = ", "),
      "\n"
    )
  }
  cat("  requirements:")
  if (length(reqs) == 0) {
    cat(" <none>\n")
  } else {
    cat("\n")
    for (req in reqs) {
      typed_args <- names(req@args)
      typed_args <- if (length(typed_args) == 0) {
        ""
      } else {
        sprintf(" args: %s", paste(typed_args, collapse = ", "))
      }
      cat(sprintf("    - %s()%s\n", req@name, typed_args))
    }
  }
  invisible(x)
}

library(S7)

local({
  Canvas <- new_class("CanvasProgressiveTest")
  Circle <- new_class(
    "CircleProgressiveTest",
    properties = list(r = class_double)
  )
  Square <- new_class(
    "SquareProgressiveTest",
    properties = list(side = class_double)
  )

  draw_on <- new_generic(
    "draw_on_progressive_test",
    c("x", "canvas"),
    function(x, canvas, position, ...) S7_dispatch()
  )

  method(draw_on, list(Circle, Canvas)) <- function(x, canvas, position, ...) {
    sprintf("circle(%s)@%s", x@r, position)
  }
  method(draw_on, list(Square, Canvas)) <- function(x, canvas, position, ...) {
    x@side
  }

  DrawableOnCanvas <- new_interface(
    "DrawableOnCanvasProgressiveTest",
    generics = list(
      draw_on = interface_requirement(
        draw_on,
        args = list(canvas = Canvas, position = class_integer),
        returns = class_character
      )
    )
  )

  circle <- Circle(r = 2)
  square <- Square(side = 3)
  canvas <- Canvas()

  expect_true(implements(Circle, DrawableOnCanvas))
  expect_true(implements(circle, DrawableOnCanvas))
  expect_equal(
    with(DrawableOnCanvas, draw_on(circle, canvas, position = 1L)),
    "circle(2)@1"
  )
  expect_equal(
    draw_on(circle, canvas, position = 1L) %::% DrawableOnCanvas,
    "circle(2)@1"
  )

  n_eval <- 0L
  make_circle <- function() {
    n_eval <<- n_eval + 1L
    circle
  }
  make_position <- function() {
    n_eval <<- n_eval + 1L
    1L
  }
  expect_equal(
    with(
      DrawableOnCanvas,
      draw_on(make_circle(), canvas, position = make_position())
    ),
    "circle(2)@1"
  )
  expect_equal(n_eval, 2L)

  local({
    draw_on <- function(...) "not the S7 generic"
    expect_equal(
      with(DrawableOnCanvas, draw_on(circle, canvas, position = 1L)),
      "circle(2)@1"
    )
  })

  expect_equal(
    with(DrawableOnCanvas, {
      render <- function(x) draw_on(x, canvas, position = 1L)
      render(circle)
    }),
    "circle(2)@1"
  )

  render_checked <- with(DrawableOnCanvas, function(x) {
    draw_on(x, canvas, position = 1L)
  })
  expect_equal(render_checked(circle), "circle(2)@1")
  expect_error(render_checked(square), "Return value")

  render_checked2 <- (function(x) draw_on(x, canvas, position = 1L)) %::%
    DrawableOnCanvas
  expect_equal(render_checked2(circle), "circle(2)@1")
  expect_error(render_checked2(square), "Return value")

  render_plain <- function(x) draw_on(x, canvas, position = 1L)
  render_checked3 <- render_plain %::% DrawableOnCanvas
  expect_equal(render_checked3(circle), "circle(2)@1")
  expect_error(render_checked3(square), "Return value")

  render_checked4 <- with(DrawableOnCanvas, render_plain)
  expect_equal(render_checked4(circle), "circle(2)@1")
  expect_error(render_checked4(square), "Return value")

  render_capturing <- local({
    pos <- 1L
    function(x) draw_on(x, canvas, position = pos)
  })
  render_checked5 <- render_capturing %::% DrawableOnCanvas
  expect_equal(render_checked5(circle), "circle(2)@1")
  expect_error(render_checked5(square), "Return value")

  expect_error(
    with(DrawableOnCanvas, draw_on(circle, canvas, position = "bad")),
    "position"
  )
  expect_error(
    with(DrawableOnCanvas, draw_on(square, canvas, position = 1L)),
    "Return value"
  )

  draw_default <- new_generic(
    "draw_default_progressive_test",
    c("x", "canvas"),
    function(x, canvas, position = 1L, ...) S7_dispatch()
  )
  method(draw_default, list(Circle, Canvas)) <- function(
    x,
    canvas,
    position = 1L,
    ...
  ) {
    sprintf("default-circle(%s)@%s", x@r, position)
  }
  DrawableDefault <- new_interface(
    "DrawableDefaultProgressiveTest",
    generics = list(
      draw_default = interface_requirement(
        draw_default,
        args = list(canvas = Canvas, position = class_integer),
        returns = class_character
      )
    )
  )
  expect_equal(
    with(DrawableDefault, draw_default(circle, canvas)),
    "default-circle(2)@1"
  )

  draw_bad_default <- new_generic(
    "draw_bad_default_progressive_test",
    c("x", "canvas"),
    function(x, canvas, position = "bad", ...) S7_dispatch()
  )
  method(draw_bad_default, list(Circle, Canvas)) <- function(
    x,
    canvas,
    position = "bad",
    ...
  ) {
    sprintf("bad-default-circle(%s)@%s", x@r, position)
  }
  DrawableBadDefault <- new_interface(
    "DrawableBadDefaultProgressiveTest",
    generics = list(
      draw_bad_default = interface_requirement(
        draw_bad_default,
        args = list(canvas = Canvas, position = class_integer),
        returns = class_character
      )
    )
  )
  expect_error(
    with(DrawableBadDefault, draw_bad_default(circle, canvas)),
    "position"
  )
  expect_error(
    draw_bad_default(circle, canvas) %::% DrawableBadDefault,
    "position"
  )

  draw_dots <- new_generic(
    "draw_dots_progressive_test",
    "x",
    function(x, ..., flag = TRUE) S7_dispatch()
  )
  method(draw_dots, Circle) <- function(x, ..., flag = TRUE) {
    list(flag = flag, dots = list(...))
  }
  DrawableDots <- new_interface(
    "DrawableDotsProgressiveTest",
    generics = list(
      draw_dots = interface_requirement(draw_dots, returns = class_list)
    )
  )
  dots_out <- with(DrawableDots, draw_dots(circle, a = 1, b = 2, flag = FALSE))
  expect_equal(dots_out$flag, FALSE)
  expect_equal(dots_out$dots, list(a = 1, b = 2))

  default_from_arg <- new_generic(
    "default_from_arg_progressive_test",
    "x",
    function(x, values, n = length(values)) S7_dispatch()
  )
  method(default_from_arg, Circle) <- function(x, values, n = length(values)) {
    n + length(values)
  }
  DefaultFromArg <- new_interface(
    "DefaultFromArgProgressiveTest",
    generics = list(
      default_from_arg = interface_requirement(
        default_from_arg,
        args = list(n = class_integer),
        returns = class_integer
      )
    )
  )
  n_values_eval <- 0L
  make_values <- function() {
    n_values_eval <<- n_values_eval + 1L
    1:3
  }
  expect_equal(
    with(DefaultFromArg, default_from_arg(circle, make_values())),
    6L
  )
  expect_equal(n_values_eval, 1L)

  Plain <- new_class("PlainProgressiveTest")
  make_bad <- new_generic("make_bad_progressive_test", "x")
  method(make_bad, Circle) <- function(x) Plain()
  MakerDrawable <- new_interface(
    "MakerDrawableProgressiveTest",
    generics = list(
      make_bad = interface_requirement(make_bad, returns = DrawableOnCanvas)
    )
  )
  expect_error(
    with(MakerDrawable, make_bad(circle)),
    "Return value does not implement"
  )

  no_position <- new_generic(
    "no_position_progressive_test",
    c("x", "canvas"),
    function(x, canvas, ...) S7_dispatch()
  )
  method(no_position, list(Circle, Canvas)) <- function(x, canvas, ...) "ok"
  BadShape <- new_interface(
    "BadShapeProgressiveTest",
    generics = list(
      no_position = interface_requirement(
        no_position,
        args = list(canvas = Canvas, position = class_integer),
        returns = class_character
      )
    )
  )
  expect_false(implements(Circle, BadShape))
  expect_equal(
    missing_requirements(Circle, BadShape)$requirement,
    "no_position"
  )

  trait_draw_on <- new_generic(
    "trait_draw_on_progressive_test",
    c("x", "canvas"),
    function(x, canvas, position, ...) S7_dispatch()
  )

  DrawableTrait <- new_trait(
    "DrawableTraitProgressiveTest",
    methods = list(
      trait_draw_on = trait_method(
        trait_draw_on,
        args = list(canvas = Canvas, position = class_integer),
        returns = class_character
      )
    )
  )

  impl_trait(
    DrawableTrait,
    Circle,
    methods = list(trait_draw_on = function(x, canvas, position, ...) {
      sprintf("trait-circle(%s)@%s", x@r, position)
    })
  )

  expect_equal(
    with(DrawableTrait, trait_draw_on(circle, canvas, position = 2L)),
    "trait-circle(2)@2"
  )
  expect_equal(
    trait_draw_on(circle, canvas, position = 2L) %::% DrawableTrait,
    "trait-circle(2)@2"
  )
  expect_error(
    with(DrawableTrait, trait_draw_on(circle, canvas, position = "bad")),
    "position"
  )
})

library(S7)

local({
  area <- new_generic("area_interface_test", "x")
  draw <- new_generic("draw_interface_test", "x")

  Circle <- new_class(
    "CircleInterfaceTest",
    properties = list(r = class_double)
  )
  Rect <- new_class(
    "RectInterfaceTest",
    properties = list(w = class_double, h = class_double)
  )

  method(area, Circle) <- function(x) pi * x@r^2
  method(draw, Circle) <- function(x) sprintf("circle(%s)", x@r)
  method(area, Rect) <- function(x) x@w * x@h

  Drawable <- new_interface(
    "DrawableInterfaceTest",
    generics = list(draw = draw)
  )
  Shape <- new_interface(
    "ShapeInterfaceTest",
    generics = list(area = area),
    parents = Drawable
  )
  DrawableAlias <- new_interface(
    "DrawableAliasInterfaceTest",
    methods = list(draw = draw)
  )

  expect_equal(names(interface_requirements(Drawable)), "draw")
  expect_equal(names(interface_requirements(DrawableAlias)), "draw")
  expect_error(
    new_interface("BadInterfaceTest", generics = list(identity)),
    "S7 generic"
  )
  expect_error(
    new_interface(
      "ConflictInterfaceTest",
      generics = list(draw = draw),
      methods = list(area = area)
    ),
    "either"
  )
  expect_true(any(grepl(
    "<S7 Go-like interface> DrawableInterfaceTest",
    capture.output(print(Drawable)),
    fixed = TRUE
  )))
  expect_true(implements(Circle, Shape))
  expect_true(implements(Circle(r = 2), Shape))
  expect_false(implements(Rect, Shape))
  expect_equal(missing_requirements(Rect, Shape)$requirement, "draw")
  expect_identical(as_interface(Circle(r = 2), Drawable)@r, 2)
  expect_error(assert_implements(Rect, Shape), "missing: draw\\(\\)")
})

#' s7contract: Contract Helpers for S7
#'
#' `s7contract` provides two experimental contract layers on top of S7:
#'
#' - Go-like structural interfaces defined by required generics.
#' - Rust-like explicit traits with default methods and associated metadata.
#'
#' The package keeps actual method dispatch inside ordinary S7 generics and uses
#' runtime checks to describe or assert conformance.
#'
#' @docType package
#' @name s7contract
"_PACKAGE"

.onLoad <- function(...) {
  S7::method(print, s7_interface) <- .print_s7_interface
  S7::method(print, s7_trait) <- .print_s7_trait
  S7::method(with, s7_interface) <- .with_s7_interface
  S7::method(with, s7_trait) <- .with_s7_trait
  S7::methods_register()
}

# Helpers for optional, progressive runtime typing of contract requirements.

.normalise_type_specs <- function(args, what = "args") {
  if (is.null(args)) {
    return(list())
  }
  if (!is.list(args)) {
    .abort("`%s` must be a named list of type specifications.", what)
  }
  if (length(args) > 0 && (is.null(names(args)) || any(names(args) == ""))) {
    .abort("`%s` must be a named list of type specifications.", what)
  }
  for (name in names(args)) {
    .check_type_spec(args[[name]], sprintf("%s$%s", what, name))
  }
  args
}

.normalise_return_spec <- function(returns) {
  if (is.null(returns)) {
    return(S7::class_any)
  }
  .check_type_spec(returns, "returns")
  returns
}

.check_type_spec <- function(spec, arg) {
  if (.is_interface(spec) || .is_trait(spec)) {
    return(invisible(spec))
  }
  if (is.null(.as_class_or_null(spec, arg = arg))) {
    .abort("`%s` must be an S7 class, S7 union, interface, or trait.", arg)
  }
  invisible(spec)
}

.spec_as_dispatch_class <- function(spec, arg) {
  if (.is_interface(spec) || .is_trait(spec)) {
    .abort(
      "`%s` is a dispatch argument and must be an S7 class or S7 union, not an interface or trait.",
      arg
    )
  }
  cls <- .as_class_or_null(spec, arg = arg)
  if (is.null(cls)) {
    .abort("`%s` must be an S7 class or S7 union for S7 dispatch.", arg)
  }
  cls
}

.value_error_label <- function(arg) {
  if (identical(arg, ".return")) "Return value" else sprintf("`%s`", arg)
}

.check_value_conforms <- function(value, spec, arg) {
  if (.is_interface(spec)) {
    assert_implements(value, spec, arg = .value_error_label(arg))
    return(invisible(value))
  }
  if (.is_trait(spec)) {
    assert_trait(value, spec, arg = .value_error_label(arg))
    return(invisible(value))
  }

  checker <- S7::new_class(
    "s7contract_value_checker",
    package = "s7contract",
    properties = list(value = spec)
  )
  tryCatch(
    checker(value = value),
    error = function(e) {
      msg <- conditionMessage(e)
      lines <- strsplit(msg, "\n", fixed = TRUE)[[1L]]
      value_line <- grep("^- @value ", lines, value = TRUE)
      if (length(value_line) > 0) {
        msg <- sub("^- @value ", "", value_line[[1L]])
      }
      .abort("%s %s", .value_error_label(arg), msg)
    }
  )
  invisible(value)
}

.check_required_formals <- function(fun, arg_names, what) {
  if (length(arg_names) == 0) {
    return(invisible(TRUE))
  }
  formals_names <- names(formals(fun))
  missing <- setdiff(arg_names, formals_names)
  if (length(missing) > 0) {
    .abort(
      "%s is missing required argument(s): %s",
      what,
      paste(missing, collapse = ", ")
    )
  }
  invisible(TRUE)
}

.requirement_signature <- function(req, target) {
  generic <- req@generic
  dispatch_args <- if (inherits(generic, "S7_generic")) {
    generic@dispatch_args
  } else {
    character()
  }
  cls <- .target_class_or_null(target, arg = "x")
  if (is.null(cls)) {
    .abort("Could not determine the class of the first dispatch argument.")
  }
  if (length(dispatch_args) <= 1L) {
    return(cls)
  }

  signature <- vector("list", length(dispatch_args))
  names(signature) <- dispatch_args
  signature[[1L]] <- cls

  for (arg in dispatch_args[-1L]) {
    if (!arg %in% names(req@args)) {
      .abort("Missing type specification for dispatch argument `%s`.", arg)
    }
    signature[[arg]] <- .spec_as_dispatch_class(req@args[[arg]], arg)
  }

  signature
}

.lookup_requirement_method <- function(req, target) {
  generic <- req@generic

  tryCatch(
    {
      signature <- .requirement_signature(req, target)
      method <- S7::method(generic, class = signature)

      .check_required_formals(
        generic,
        names(req@args),
        sprintf("Generic `%s()`", req@name)
      )
      .check_required_formals(
        method,
        names(req@args),
        sprintf("Method `%s()`", req@name)
      )

      list(ok = TRUE, method = method, error = NULL)
    },
    error = function(e) {
      list(ok = FALSE, method = NULL, error = e)
    }
  )
}


.bind_checked_arg <- function(call, arg, value, eval_env, prefix) {
  nm <- sprintf(".%s_%s", prefix, arg)
  assign(nm, value, envir = eval_env)
  assign(arg, value, envir = eval_env)
  call[[arg]] <- as.name(nm)
  call
}

.bind_delayed_arg <- function(call, arg, expr, eval_env, source_env, prefix) {
  nm <- sprintf(".%s_%s", prefix, arg)
  force(expr)
  force(eval_env)
  force(source_env)
  force(nm)

  delayedAssign(nm, eval(expr, envir = source_env), assign.env = eval_env)
  delayedAssign(arg, get(nm, envir = eval_env), assign.env = eval_env)
  call[[arg]] <- as.name(nm)
  call
}

.is_missing_default <- function(x) {
  identical(x, quote(expr = ))
}

.default_typed_arg <- function(generic, method, arg, eval_env) {
  for (fun in list(method, generic)) {
    fml <- formals(fun)
    if (arg %in% names(fml) && !.is_missing_default(fml[[arg]])) {
      return(eval(fml[[arg]], envir = eval_env))
    }
  }
  .abort(
    "Call is missing typed argument `%s` and no default could be evaluated.",
    arg
  )
}

.generic_bind_names <- function(req) {
  unique(c(req@name, .generic_label(req@generic)))
}

.checked_generic_call <- function(contract, req, call, env, trait = FALSE) {
  generic <- req@generic
  matched <- tryCatch(
    match.call(definition = generic, call = call, expand.dots = TRUE),
    error = function(e) {
      .abort(
        "Could not match call to `%s()`: %s",
        req@name,
        conditionMessage(e)
      )
    }
  )

  eval_env <- new.env(parent = env)
  assign(".s7contract_generic", generic, envir = eval_env)
  matched[[1L]] <- as.name(".s7contract_generic")

  formal_args <- names(formals(generic))
  supplied_args <- intersect(names(matched)[-1L], formal_args)
  supplied_args <- setdiff(supplied_args[nzchar(supplied_args)], "...")
  for (arg in supplied_args) {
    matched <- .bind_delayed_arg(
      matched,
      arg,
      matched[[arg]],
      eval_env,
      env,
      "arg"
    )
  }

  dispatch_args <- if (inherits(generic, "S7_generic")) {
    generic@dispatch_args
  } else {
    names(formals(generic))[1L]
  }
  first_arg <- dispatch_args[[1L]]
  if (!first_arg %in% names(matched)) {
    .abort("Call is missing first dispatch argument `%s`.", first_arg)
  }
  first_value <- eval(matched[[first_arg]], envir = eval_env)
  matched <- .bind_checked_arg(matched, first_arg, first_value, eval_env, "arg")

  if (trait) {
    assert_trait(first_value, contract, arg = first_arg)
  } else {
    assert_implements(first_value, contract, arg = first_arg)
  }

  found <- .lookup_requirement_method(req, first_value)
  if (!found$ok) {
    .abort("%s", conditionMessage(found$error))
  }

  for (arg in names(req@args)) {
    value <- if (arg %in% names(matched)) {
      eval(matched[[arg]], envir = eval_env)
    } else {
      .default_typed_arg(generic, found$method, arg, eval_env)
    }
    .check_value_conforms(value, req@args[[arg]], arg)
    matched <- .bind_checked_arg(matched, arg, value, eval_env, "arg")
  }

  out <- eval(matched, envir = eval_env)
  .check_value_conforms(out, req@returns, ".return")
  out
}

.make_checked_generic <- function(contract, req, trait = FALSE) {
  force(contract)
  force(req)
  force(trait)
  function(...) {
    .checked_generic_call(
      contract,
      req,
      sys.call(),
      parent.frame(),
      trait = trait
    )
  }
}

.contract_mask <- function(contract, env, trait = FALSE) {
  reqs <- if (trait) {
    trait_methods(contract, inherited = TRUE)
  } else {
    interface_requirements(contract, inherited = TRUE)
  }
  mask <- new.env(parent = env)

  for (req in reqs) {
    wrapper <- .make_checked_generic(contract, req, trait = trait)
    for (name in .generic_bind_names(req)) {
      assign(name, wrapper, envir = mask)
    }
  }

  mask
}

.rebind_contract_function <- function(fun, contract, mask, trait = FALSE) {
  if (!identical(typeof(fun), "closure") || identical(environment(fun), mask)) {
    return(fun)
  }

  function_mask <- .contract_mask(contract, environment(fun), trait = trait)
  environment(fun) <- function_mask
  fun
}

.with_contract <- function(contract, expr, env, trait = FALSE) {
  mask <- .contract_mask(contract, env, trait = trait)
  out <- eval(expr, envir = mask)
  .rebind_contract_function(out, contract, mask, trait = trait)
}

.with_s7_interface <- function(data, expr, ...) {
  .with_contract(data, substitute(expr), parent.frame(), trait = FALSE)
}

.with_s7_trait <- function(data, expr, ...) {
  .with_contract(data, substitute(expr), parent.frame(), trait = TRUE)
}

#' Evaluate an S7 call under an interface or trait contract
#'
#' `with(contract, expr)` and `expr %::% contract` evaluate `expr` in a
#' contract mask. Required generics are shadowed by checking wrappers, so calls
#' to those generics use normal S7 dispatch while checking the optional argument
#' and return specifications stored in an interface requirement or trait method.
#'
#' @param expr An expression evaluated in a contract mask. Calls to generics
#'   named in the contract are checked.
#' @param contract An interface created by [new_interface()] or a trait created
#'   by [new_trait()].
#' @return The value of `expr`, after any optional return check.
#' @aliases contract_syntax
#' @examples
#' local({
#'   draw <- S7::new_generic("draw", "x", function(x, color) {
#'     S7::S7_dispatch()
#'   })
#'   Circle <- S7::new_class("TypedCircle", properties = list(r = S7::class_double))
#'   S7::method(draw, Circle) <- function(x, color) paste(color, x@r)
#'   Drawable <- new_interface(
#'     "TypedDrawable",
#'     generics = list(draw = interface_requirement(
#'       draw,
#'       args = list(color = S7::class_character),
#'       returns = S7::class_character
#'     ))
#'   )
#'   with(Drawable, draw(Circle(r = 2), color = "red"))
#'   checked_draw <- with(Drawable, function(x) draw(x, color = "red"))
#'   checked_draw(Circle(r = 2))
#'   draw(Circle(r = 2), color = "red") %::% Drawable
#' })
#' @export
`%::%` <- function(expr, contract) {
  contract <- eval(substitute(contract), envir = parent.frame())
  if (.is_interface(contract)) {
    return(.with_contract(
      contract,
      substitute(expr),
      parent.frame(),
      trait = FALSE
    ))
  }
  if (.is_trait(contract)) {
    return(.with_contract(
      contract,
      substitute(expr),
      parent.frame(),
      trait = TRUE
    ))
  }
  .abort("Right-hand side of `%::%` must be an interface or trait.")
}

---
title: "Bioinformatics Containers as Interfaces"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Bioinformatics Containers as Interfaces}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(S7)
library(s7contract)
```

## Introduction

Bioinformatics packages often exchange rich containers rather than plain
matrices.  Bioconductor's `SummarizedExperiment`, for example, keeps assays,
feature metadata, sample metadata, names, and validity rules synchronized.  This
vignette does not try to rebuild that class.  It uses a small toy container to
show what an interface can and cannot express.

The useful idea is an adapter boundary.  A downstream function might not need a
specific container class.  It might only need to retrieve one assay matrix, or it
might need assay names plus feature and sample names.  That small behavior can be
written as an S7 interface at the point where the downstream function consumes
it.

## Background

A class and an interface answer different design questions.  A class describes
representation and invariants: where the assays live, how metadata is stored,
and what must be true after construction or subsetting.  An interface describes
behavior: which operations a consumer may call.

This distinction matters for bioinformatics.  A behavioral interface can make
small examples, tests, and adapters easier to write, but it does not replace a
well-established interoperability class.  It cannot enforce genome builds,
biological interpretation, delayed computation, or the full set of conventions
used by Bioconductor containers.

## A toy assay container

The example class stores a named list of assays plus row and column metadata.
The validator only checks the dimensions needed below.

```{r toy-container}
MiniSummarizedExperiment <- new_class(
  "MiniSummarizedExperiment",
  properties = list(
    assays = class_list,
    row_data = class_data.frame,
    col_data = class_data.frame
  ),
  validator = function(self) {
    if (length(self@assays) == 0) {
      return("@assays must contain at least one matrix")
    }

    dims <- lapply(self@assays, dim)
    if (any(vapply(dims, is.null, logical(1)))) {
      return("every assay must be matrix-like")
    }

    first_dim <- dims[[1]]
    same_dim <- vapply(dims, identical, logical(1), first_dim)
    if (!all(same_dim)) {
      return("all assays must have the same dimensions")
    }

    if (nrow(self@row_data) != first_dim[[1]]) {
      return("@row_data must have one row per assay feature")
    }
    if (nrow(self@col_data) != first_dim[[2]]) {
      return("@col_data must have one row per assay sample")
    }
  }
)

counts <- matrix(
  c(10, 0, 3, 4, 12, 8),
  nrow = 3,
  dimnames = list(c("geneA", "geneB", "geneC"), c("sample1", "sample2"))
)

mini <- MiniSummarizedExperiment(
  assays = list(counts = counts, logcounts = log1p(counts)),
  row_data = data.frame(gc = c(0.42, 0.51, 0.37), row.names = rownames(counts)),
  col_data = data.frame(condition = c("control", "treated"), row.names = colnames(counts))
)
```

## Operations and a consumer-owned interface

Adapters expose behavior through ordinary S7 generics.  The toy container below
supports assay names, feature names, sample names, and assay lookup, but a
consumer should only require the operations it actually uses.

```{r assay-operations}
assay_names <- new_generic("assay_names", "x")
feature_names <- new_generic("feature_names", "x")
sample_names <- new_generic("sample_names", "x")
assay_matrix <- new_generic("assay_matrix", "x")

method(assay_names, MiniSummarizedExperiment) <- function(x) names(x@assays)
method(feature_names, MiniSummarizedExperiment) <- function(x) rownames(x@assays[[1]])
method(sample_names, MiniSummarizedExperiment) <- function(x) colnames(x@assays[[1]])
method(assay_matrix, MiniSummarizedExperiment) <- function(x, name = assay_names(x)[[1]]) {
  x@assays[[name]]
}

assay_names(mini)
sample_names(mini)
assay_matrix(mini, "counts")[, "sample1"]
```

The interface belongs at the point of use.  A library-size calculation does not
need feature metadata or sample metadata; it only needs to retrieve one assay
matrix.  This mirrors the Go pattern `func Takes(db Database) error`: accept the
small protocol the function needs, not a concrete database or a giant package
interface.

```{r assay-consumer}
LibrarySizeInput <- new_interface(
  "LibrarySizeInput",
  generics = list(assay_matrix = assay_matrix)
)

library_size <- function(x, assay = "counts") {
  assert_implements(x, LibrarySizeInput)
  mat <- assay_matrix(x, assay)
  colSums(mat)
}

implements(mini, LibrarySizeInput)
library_size(mini)
```

The payoff is testing.  A unit test does not need to construct a realistic
`MiniSummarizedExperiment` or a full Bioconductor object.  It can provide a tiny
mock that implements exactly the consumer-owned protocol.

```{r assay-mock}
MockAssays <- new_class("MockAssays", properties = list(assays = class_list))

method(assay_matrix, MockAssays) <- function(x, name = "counts") {
  x@assays[[name]]
}

mock_counts <- matrix(
  c(1, 2, 3, 4),
  nrow = 2,
  dimnames = list(c("geneA", "geneB"), c("sampleA", "sampleB"))
)
mock <- MockAssays(assays = list(counts = mock_counts))

implements(mock, LibrarySizeInput)
library_size(mock)
```

This is the productive use case.  A package can write against a small protocol,
return an ordinary vector, and let separate adapters provide methods for
concrete containers.

## When an explicit trait helps

A trait is useful when structural compatibility is not enough.  Here the trait
records an explicit implementation and stores an associated constant describing
assay orientation.

```{r assay-trait, message = FALSE, warning = FALSE}
ExperimentLike <- new_trait(
  "ExperimentLike",
  methods = list(
    assay_names = trait_method(assay_names),
    feature_names = trait_method(feature_names),
    sample_names = trait_method(sample_names),
    assay_matrix = trait_method(assay_matrix)
  ),
  assoc_consts = c("ASSAY_ORIENTATION")
)

impl_trait(
  ExperimentLike,
  MiniSummarizedExperiment,
  methods = list(
    assay_names = function(x) names(x@assays),
    feature_names = function(x) rownames(x@assays[[1]]),
    sample_names = function(x) colnames(x@assays[[1]]),
    assay_matrix = function(x, name = assay_names(x)[[1]]) x@assays[[name]]
  ),
  assoc_consts = list(ASSAY_ORIENTATION = "features_by_samples"),
  replace = TRUE
)

has_trait(mini, ExperimentLike)
trait_assoc_const(ExperimentLike, mini, "ASSAY_ORIENTATION")
```

## Design cautions

It would be a mistake to define one large interface or trait that tries to cover
every bioinformatics object.  Assay matrices, genomic ranges, variant calls, and
single-cell objects have different invariants and different performance needs.
If a consumer only needs `assay_matrix()`, do not make it depend on feature
metadata, sample metadata, genome ranges, and delayed computation as well.  Small
interfaces are easier to satisfy correctly and easier to test.

It would also be a mistake to claim that an interface proves biological
correctness.  Method availability does not prove that samples are comparable,
that row ranges use the same genome build, or that an assay transform is
appropriate for a downstream model.  Those checks should remain explicit and
domain-specific.

The narrow conclusion is useful enough: interfaces can mimic a small behavioral
slice of a class such as `SummarizedExperiment`, but they should not replace the
class or its ecosystem.

## References

- The Bioconductor `SummarizedExperiment` package: <https://bioconductor.org/packages/SummarizedExperiment/>.
- Morgan et al. (2023), "Orchestrating high-throughput genomic analysis with Bioconductor": <https://bioconductor.org/help/publications/>.
- The S7 package documentation: <https://rconsortium.github.io/S7/>.
- Chewxy, "How To Use Go Interfaces": <https://blog.chewxy.com/2018/03/18/golang-interfaces/>.
- The `s7contract` interface and trait vignette in this package.

---
title: "Go-Like Interfaces and Rust-Like Traits on S7"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Go-Like Interfaces and Rust-Like Traits on S7}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(S7)
library(s7contract)
```

## Introduction

`s7contract` is a small experiment: can interface and trait ideas be expressed
on top of S7 without replacing S7's method system?  The package answers yes, but
with an important constraint.  S7 still owns dispatch.  `s7contract` only records
and checks contracts around ordinary S7 generics.

The most natural layer is a Go-like structural interface.  In S7, operations are
already ordinary functions such as `draw(x)` or `area(x)`, and methods are
registered for classes.  A structural interface can therefore be just a named set
of required generics.

A Rust-like trait is also possible, but it needs an explicit registry.  That
extra machinery is useful for default methods and associated metadata, but it is
less native to S7 because R does not have Rust's compile-time trait bounds,
coherence rules, or associated type checker.

## Background

S7 is a functional object-oriented system: methods belong to generic functions,
not to objects.  The call is `generic(object, ...)`, not `object$generic(...)`.
That makes S7 close in spirit to protocols defined by behavior.

Go interfaces are structural: a basic interface describes required methods, and
a type satisfies the interface when it has those methods.  Go style also favors
small interfaces defined at the point of use: do not define an interface beside a
single implementation merely to make that implementation conform.  In S7 terms,
define classes, generics, and methods where the behavior lives, then let the
consumer define the protocol it accepts.  Rust traits are nominal and explicit:
an implementation is declared for a type, and traits may also contain defaults
and associated items.  `s7contract` maps these ideas to S7 as follows.

```text
S7 generic       operation, e.g. area(x)
S7 method        implementation for a class
Go-like interface set of required S7 generics
Rust-like trait  explicit implementation record plus S7 methods
```

This also clarifies what kind of "type" an interface defines.  An S7 class is a
nominal representation type: it says how an object is constructed and validated.
An interface is a behavioral or protocol type: it says what operations must be
available.  Both are useful, but they answer different questions.

Packages such as `lambda.r` explore a different functional-programming route in
R, with pattern-matching-style function clauses, guards, and optional type
constraints.  `s7contract` stays closer to S7: it does not create a new function
clause language.  By default it checks whether S7 can find methods for required
generics, or whether an explicit trait implementation has been registered.
Optional argument and return specifications can be checked when evaluating a
call with `with()` or `%::%`.

## A structural interface

The classic drawing example remains useful because the behavior is visible.  A
`Drawable` object is anything for which S7 can find a `draw()` method.  In real
packages, this interface would usually live near the consumer that needs to draw
things, not necessarily in the package that defines `Circle`.

```{r structural-interface}
area <- new_generic("area", "x")
draw <- new_generic("draw", "x")

Circle <- new_class("Circle", properties = list(r = class_double))
Rect <- new_class("Rect", properties = list(w = class_double, h = class_double))

method(area, Circle) <- function(x) pi * x@r^2
method(draw, Circle) <- function(x) sprintf("circle(r = %s)", x@r)
method(area, Rect) <- function(x) x@w * x@h

Drawable <- new_interface("Drawable", generics = list(draw = draw))
Shape <- new_interface("Shape", generics = list(area = area), parents = Drawable)

implements(Circle, Shape)
implements(Rect, Shape)
missing_requirements(Rect, Shape)
```

A consumer keeps ordinary S7 style.  The assertion documents the expected
behavior; the actual call is still normal dispatch through `draw(x)`.

```{r structural-consumer}
render <- function(x) {
  assert_implements(x, Drawable)
  draw(x)
}

render(Circle(r = 2))
```

The same boundary is convenient in tests.  A mock only needs the behavior the
consumer asks for.

```{r structural-mock}
MockDrawable <- new_class("MockDrawable")
method(draw, MockDrawable) <- function(x) "mock drawing"

render(MockDrawable())
```

This is the main reason structural interfaces fit S7 well.  They add a small
runtime check around a dispatch model that S7 already has.  The practical API
shape is the Go maxim adapted to R: accept objects that satisfy a small protocol;
return ordinary, concrete R or S7 values.

## When the whole protocol is a class family

Some R APIs intentionally define a large protocol up front.  DBI is the useful
example: it is not just one consumer-local interface, but a package-level
standard built around nominal connection, driver, and result classes plus many
generic functions.  That fits the "abstract data type" exception to the
point-of-use rule.

In S7, the analogous design is a class family for identity and representation,
plus generics for behavior.  Consumers can still depend on a smaller interface
when they only need part of the protocol.

```{r dbi-like-class-family}
DatabaseConnection <- new_class("DatabaseConnection", abstract = TRUE)
MemoryConnection <- new_class(
  "MemoryConnection",
  parent = DatabaseConnection,
  properties = list(tables = class_list)
)

db_tables <- new_generic("db_tables", "con")
db_read_table <- new_generic(
  "db_read_table",
  "con",
  function(con, name) S7_dispatch()
)

method(db_tables, MemoryConnection) <- function(con) names(con@tables)
method(db_read_table, MemoryConnection) <- function(con, name) con@tables[[name]]

TableReader <- new_interface(
  "TableReader",
  generics = list(
    db_tables = interface_requirement(db_tables, returns = class_character),
    db_read_table = interface_requirement(
      db_read_table,
      args = list(name = class_character),
      returns = class_data.frame
    )
  )
)

first_table <- function(con) {
  assert_implements(con, TableReader)
  db_read_table(con, db_tables(con)[[1]])
}

con <- MemoryConnection(tables = list(iris = head(iris, 2)))
first_table(con)
```

The class says "this object is a database connection".  The interface says
"this consumer needs table-reading behavior".  A full DBI-like package may own
the broad class family; ordinary downstream functions should still prefer the
smallest protocol they use.

## Progressive argument and return checks

Interface requirements can optionally carry argument and return specifications.
The default is permissive: unspecified arguments are not checked, and the return
specification defaults to `S7::class_any`.  When specifications are present,
expressions can be evaluated in a contract mask with either `with()` or the
lambda.r-style `%::%` operator.  Calls to required generics inside that
expression are checked.

```{r progressive-interface}
Canvas <- new_class("Canvas")

draw_on <- new_generic(
  "draw_on",
  c("x", "canvas"),
  function(x, canvas, position, ...) S7_dispatch()
)

method(draw_on, list(Circle, Canvas)) <- function(x, canvas, position, ...) {
  sprintf("circle(r = %s) at %s", x@r, position)
}

DrawableOnCanvas <- new_interface(
  "DrawableOnCanvas",
  generics = list(
    draw_on = interface_requirement(
      draw_on,
      args = list(canvas = Canvas, position = class_integer),
      returns = class_character
    )
  )
)

canvas <- Canvas()
circle <- Circle(r = 2)

implements(Circle, DrawableOnCanvas)
with(DrawableOnCanvas, draw_on(circle, canvas, position = 1L))
draw_on(circle, canvas, position = 1L) %::% DrawableOnCanvas

checked_draw <- with(DrawableOnCanvas, {
  function(x) draw_on(x, canvas, position = 1L)
})
checked_draw(circle)
```

A method can satisfy the S7 method shape but still return the wrong kind of
value.  The checked call, including a function returned from `with()`, catches
that after ordinary S7 dispatch has run.

```{r progressive-return-failure}
BadCircle <- new_class("BadCircle", properties = list(r = class_double))
method(draw_on, list(BadCircle, Canvas)) <- function(x, canvas, position, ...) {
  x@r
}

tryCatch(
  with(DrawableOnCanvas, draw_on(BadCircle(r = 2), canvas, position = 1L)),
  error = function(e) conditionMessage(e)
)

tryCatch(
  checked_draw(BadCircle(r = 2)),
  error = function(e) conditionMessage(e)
)
```

The input checks use S7 classes and S7 multiple dispatch.  In this example,
`canvas` is also a dispatch argument, so `implements(Circle, DrawableOnCanvas)`
asks S7 for a `draw_on(<Circle>, <Canvas>)` method.  The return value can only be
checked after the call has run, which is why `with()` and `%::%` are useful.

## Number-like behavior

A general `Number` interface is tempting, but it should be treated carefully.
Base R arithmetic includes vectorization, recycling, missing values, attributes,
and binary operations.  A small number-like protocol is more honest: it says
only which operations a particular consumer needs.

```{r number-interface}
num_zero <- new_generic("num_zero", "x")
num_add <- new_generic("num_add", "x")
num_scale <- new_generic("num_scale", "x")

NumberLike <- new_interface(
  "NumberLike",
  generics = list(
    zero = num_zero,
    add = num_add,
    scale = num_scale
  )
)

method(num_zero, class_double) <- function(x) 0
method(num_add, class_double) <- function(x, y) x + y
method(num_scale, class_double) <- function(x, k) x * k

implements(class_double, NumberLike)
num_add(10, 5)
num_scale(10, 0.5)
```

This interface does not prove mathematical laws such as associativity or an
identity element.  It only says that these operations are present.  If those laws
matter, they should be described in the documentation and tested with examples
that are specific to the domain.

## Vector-like behavior

A vector-like contract is often more practical.  Many algorithms only need a
length, a way to slice, and a way to expose values.

```{r vector-interface}
vec_length <- new_generic("vec_length", "x")
vec_slice <- new_generic("vec_slice", "x")
vec_values <- new_generic("vec_values", "x")

VectorLike <- new_interface(
  "VectorLike",
  generics = list(
    length = vec_length,
    slice = vec_slice,
    values = vec_values
  )
)

ReadDepth <- new_class(
  "ReadDepth",
  properties = list(
    position = class_integer,
    depth = class_double
  ),
  validator = function(self) {
    if (length(self@position) != length(self@depth)) {
      "@position and @depth must have the same length"
    }
  }
)

method(vec_length, ReadDepth) <- function(x) length(x@depth)
method(vec_slice, ReadDepth) <- function(x, i) {
  ReadDepth(position = x@position[i], depth = x@depth[i])
}
method(vec_values, ReadDepth) <- function(x) x@depth

coverage <- ReadDepth(
  position = 1:5,
  depth = c(12, 15, 9, 20, 17)
)

implements(coverage, VectorLike)
vec_values(vec_slice(coverage, 2:4))
```

A function can depend on this small protocol without knowing how the object is
represented internally.

```{r vector-consumer}
window_mean <- function(x, i) {
  assert_implements(x, VectorLike)
  mean(vec_values(vec_slice(x, i)))
}

window_mean(coverage, 2:4)
```

This kind of interface is best used at package boundaries.  It is not meant to
replace base vectors, S7 classes, or mature vector frameworks; it names the small
piece of behavior a consumer needs.

## An explicit trait

A Rust-like trait adds nominal intent.  A class may have the right methods
structurally, but it does not have the trait until `impl_trait()` records that
implementation.

```{r trait}
perimeter <- new_generic("perimeter", "x")

Measurable <- new_trait(
  "Measurable",
  methods = list(
    area = trait_method(area),
    perimeter = trait_method(perimeter, default = function(x) NA_real_)
  ),
  assoc_consts = c("UNITS")
)

impl_trait(
  Measurable,
  Circle,
  methods = list(area = function(x) pi * x@r^2),
  assoc_consts = list(UNITS = "unitless"),
  replace = TRUE
)

has_trait(Circle, Measurable)
trait_call(Measurable, "area", Circle(r = 2))
trait_call(Measurable, "perimeter", Circle(r = 2))
trait_assoc_const(Measurable, Circle, "UNITS")
```

The useful distinction is intent.  A structural interface asks whether operations
are available.  An explicit trait asks whether a package author has declared a
class to implement a named contract.  The trait layer can also store associated
metadata such as `UNITS`, which is awkward in a purely structural interface.

## A Haskell-style dictionary object

Haskell type classes are often explained as dictionaries: a `Monad m`
constraint is operationally evidence that `m` has `pure` and `bind`.  R can
model that idea directly because functions are first-class values and S7 can
validate function-valued properties.

The example below defines a tiny `Maybe` algebraic data type, then stores its
monad operations in an S7 dictionary object.  The `s7contract` interface checks
that the dictionary exposes the operations a consumer expects.

```{r monad-dictionary}
Maybe <- new_class("Maybe", abstract = TRUE)
Nothing <- new_class("Nothing", parent = Maybe)
Just <- new_class("Just", parent = Maybe, properties = list(value = class_any))

MonadDict <- new_class(
  "MonadDict",
  properties = list(
    name = class_character,
    pure = class_function,
    bind = class_function
  )
)

dict_pure <- new_generic("dict_pure", "x")
dict_bind <- new_generic("dict_bind", "x")

MonadDictionary <- new_interface(
  "MonadDictionary",
  generics = list(
    pure = dict_pure,
    bind = dict_bind
  )
)

method(dict_pure, MonadDict) <- function(x, value) {
  (x@pure)(value)
}
method(dict_bind, MonadDict) <- function(x, mx, f) {
  (x@bind)(mx, f)
}

MaybeMonad <- MonadDict(
  name = "Maybe",
  pure = function(value) Just(value = value),
  bind = function(mx, f) {
    if (S7_inherits(mx, Nothing)) {
      Nothing()
    } else {
      f(mx@value)
    }
  }
)

implements(MaybeMonad, MonadDictionary)
```

Now the dictionary can be passed around as an ordinary R object.

```{r monad-dictionary-use}
dict_bind(
  MaybeMonad,
  Just(value = 2),
  function(x) dict_pure(MaybeMonad, x + 1)
)

dict_bind(
  MaybeMonad,
  Nothing(),
  function(x) dict_pure(MaybeMonad, x + 1)
)
```

The interface checks operation availability.  The monad laws are semantic
properties, so they belong in tests.  A small law check can still be written in
plain R.

```{r monad-laws}
maybe_equal <- function(x, y) {
  if (S7_inherits(x, Nothing) && S7_inherits(y, Nothing)) {
    return(TRUE)
  }
  if (S7_inherits(x, Just) && S7_inherits(y, Just)) {
    return(identical(x@value, y@value))
  }
  FALSE
}

f <- function(x) dict_pure(MaybeMonad, x + 1)
g <- function(x) dict_pure(MaybeMonad, x * 2)
mx <- Just(value = 10)

c(
  left_identity = maybe_equal(
    dict_bind(MaybeMonad, dict_pure(MaybeMonad, 10), f),
    f(10)
  ),
  right_identity = maybe_equal(
    dict_bind(MaybeMonad, mx, function(x) dict_pure(MaybeMonad, x)),
    mx
  ),
  associativity = maybe_equal(
    dict_bind(MaybeMonad, dict_bind(MaybeMonad, mx, f), g),
    dict_bind(MaybeMonad, mx, function(x) dict_bind(MaybeMonad, f(x), g))
  )
)
```

This is not Haskell's static kind system.  It is a concrete R/S7 encoding of the
same operational idea: a type-class instance can be represented as a runtime
object containing functions, and an interface can state which functions must be
available.

## Which feels more natural?

For functional OOP in S7, Go-like structural interfaces are the default fit.  S7
already makes generic functions the center of dispatch, so an interface as a set
of required generics is small and idiomatic.

Rust-like traits are heavier but useful when accidental compatibility would be a
problem.  They make sense for plugin systems, adapters, or domain protocols where
a package should explicitly claim conformance and provide metadata or defaults.

The practical rule is simple: start with a structural interface when the consumer
only needs behavior; use a trait when the declaration itself carries meaning.

## References

- The S7 package documentation: <https://rconsortium.github.io/S7/>.
- The Go specification, especially interface types: <https://go.dev/ref/spec#Interface_types>.
- Chewxy, "How To Use Go Interfaces": <https://blog.chewxy.com/2018/03/18/golang-interfaces/>.
- The Rust book chapter on traits: <https://doc.rust-lang.org/book/ch10-02-traits.html>.
- The Rust reference chapter on traits: <https://doc.rust-lang.org/reference/items/traits.html>.
- The `lambda.r` package on CRAN: <https://cran.r-project.org/package=lambda.r>.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interface.R
\name{interface_requirements}
\alias{interface_requirements}
\alias{interface_report}
\alias{missing_requirements}
\alias{implements}
\alias{assert_implements}
\alias{as_interface}
\title{Inspect or check a Go-like structural interface}
\usage{
interface_requirements(interface, inherited = TRUE)

interface_report(x, interface)

missing_requirements(x, interface)

implements(x, interface)

assert_implements(x, interface, arg = deparse(substitute(x)))

as_interface(x, interface)
}
\arguments{
\item{interface}{An interface created by \code{new_interface()}.}

\item{inherited}{Include inherited requirements from parent interfaces?}

\item{x}{An object, or an S7 class/base class wrapper.}

\item{arg}{Name to use in error messages.}
}
\value{
\code{interface_requirements()} returns a named list of
\code{interface_requirement()} objects. \code{interface_report()} and
\code{missing_requirements()} return data frames. \code{implements()} returns a single
logical value. \code{assert_implements()} and \code{as_interface()} return \code{x},
unchanged.
}
\description{
Inspect or check a Go-like structural interface
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/progressive.R
\name{\%::\%}
\alias{\%::\%}
\alias{contract_syntax}
\title{Evaluate an S7 call under an interface or trait contract}
\usage{
expr \%::\% contract
}
\arguments{
\item{expr}{An expression evaluated in a contract mask. Calls to generics
named in the contract are checked.}

\item{contract}{An interface created by \code{\link[=new_interface]{new_interface()}} or a trait created
by \code{\link[=new_trait]{new_trait()}}.}
}
\value{
The value of \code{expr}, after any optional return check.
}
\description{
\code{with(contract, expr)} and \code{expr \%::\% contract} evaluate \code{expr} in a
contract mask. Required generics are shadowed by checking wrappers, so calls
to those generics use normal S7 dispatch while checking the optional argument
and return specifications stored in an interface requirement or trait method.
}
\examples{
local({
  draw <- S7::new_generic("draw", "x", function(x, color) {
    S7::S7_dispatch()
  })
  Circle <- S7::new_class("TypedCircle", properties = list(r = S7::class_double))
  S7::method(draw, Circle) <- function(x, color) paste(color, x@r)
  Drawable <- new_interface(
    "TypedDrawable",
    generics = list(draw = interface_requirement(
      draw,
      args = list(color = S7::class_character),
      returns = S7::class_character
    ))
  )
  with(Drawable, draw(Circle(r = 2), color = "red"))
  checked_draw <- with(Drawable, function(x) draw(x, color = "red"))
  checked_draw(Circle(r = 2))
  draw(Circle(r = 2), color = "red") \%::\% Drawable
})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/trait.R
\name{new_trait}
\alias{new_trait}
\alias{trait_method}
\title{Build a Rust-like explicit trait on top of S7}
\usage{
new_trait(
  name,
  methods = list(),
  parents = list(),
  assoc_types = character(),
  assoc_consts = list(),
  package = NULL
)

trait_method(
  generic,
  default = NULL,
  name = NULL,
  args = list(),
  returns = S7::class_any
)
}
\arguments{
\item{name}{For \code{new_trait()}, the trait name. For \code{trait_method()}, the
method name; it defaults to the generic name when omitted.}

\item{methods}{For \code{new_trait()}, a named list of S7 generics or
\code{trait_method()} objects.}

\item{parents}{Optional trait or list of supertraits.}

\item{assoc_types}{Required associated type names, or a named list of default
associated type values.}

\item{assoc_consts}{Required associated constant names, or a named list of
default constant values.}

\item{package}{Optional package name used only for display.}

\item{generic}{An S7 generic function.}

\item{default}{Optional default implementation. If supplied, \code{impl_trait()}
uses it when a class does not provide an override for that method.}

\item{args}{Optional named list of S7 classes, interfaces, or traits for
runtime argument checking with \code{with()} or \verb{\%::\%}. Dispatch arguments other
than the first can use S7 classes or unions to refine multiple-dispatch
requirements.}

\item{returns}{Optional S7 class, interface, or trait for runtime return
checking with \code{with()} or \verb{\%::\%}; defaults to \code{S7::class_any}.}
}
\value{
\code{new_trait()} returns an S7 object of class \code{s7_trait}.
\code{trait_method()} returns an S7 object of class \code{s7_trait_method}.
}
\description{
\code{new_trait()} adds a nominal contract registry on top of S7 dispatch. A class
only has the trait after \code{impl_trait()} records the implementation, even if
compatible S7 methods already exist.
}
\details{
This makes default methods and associated metadata practical, but the result
remains a runtime R abstraction. It does not emulate Rust's compile-time
trait bounds, coherence, orphan rules, or type-checked associated types.
}
\examples{
local({
  area <- S7::new_generic("area", "x")
  perimeter <- S7::new_generic("perimeter", "x")

  Circle <- S7::new_class(
    "Circle",
    properties = list(r = S7::class_double)
  )

  Measurable <- new_trait(
    "Measurable",
    methods = list(
      area = trait_method(area),
      perimeter = trait_method(perimeter, default = function(x) NA_real_)
    ),
    assoc_consts = c("UNITS")
  )

  impl_trait(
    Measurable,
    Circle,
    methods = list(area = function(x) pi * x@r^2),
    assoc_consts = list(UNITS = "unitless")
  )

  has_trait(Circle, Measurable)
  trait_call(Measurable, "area", Circle(r = 2))
  trait_assoc_const(Measurable, Circle, "UNITS")
})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interface.R
\name{new_interface}
\alias{new_interface}
\alias{interface_requirement}
\title{Build a Go-like structural interface on top of S7}
\usage{
new_interface(
  name,
  generics = list(),
  parents = list(),
  package = NULL,
  methods = NULL
)

interface_requirement(
  generic,
  name = NULL,
  args = list(),
  returns = S7::class_any
)
}
\arguments{
\item{name}{For \code{new_interface()}, the interface name. For
\code{interface_requirement()}, the requirement name; it defaults to the generic
name when omitted.}

\item{generics}{For \code{new_interface()}, a named list of S7 generics or
\code{interface_requirement()} objects. These are generic functions because S7
methods are registered separately on generics.}

\item{parents}{Optional interface or list of interfaces to embed.}

\item{package}{Optional package name used only for display.}

\item{methods}{Compatibility alias for \code{generics}.}

\item{generic}{An S7 generic function.}

\item{args}{Optional named list of S7 classes, interfaces, or traits for
runtime argument checking with \code{with()} or \verb{\%::\%}. Arguments named in
\code{args} are also checked against generic and method formals during
conformance checks. Dispatch arguments other than the first can use S7
classes or unions to refine multiple-dispatch requirements.}

\item{returns}{Optional S7 class, interface, or trait for runtime return
checking with \code{with()} or \verb{\%::\%}; defaults to \code{S7::class_any}.}
}
\value{
\code{new_interface()} returns an S7 object of class
\code{s7_interface}. \code{interface_requirement()} returns an S7 object of class
\code{s7_interface_requirement}.
}
\description{
\code{new_interface()} models the method-list part of Go interfaces as a list of
required S7 generics. An interface is just a named set of required generics,
and a class or object satisfies it when S7 can find a method for every
required generic.
}
\details{
This deliberately mirrors Go's basic interfaces defined only by methods. The
intended style is to define small interfaces at the point where consuming code
needs a behavior, not beside a single concrete implementation. Define S7
classes, generics, and methods normally; then let consumers name the protocol
they accept. Up-front interfaces can still be useful for deliberate package
protocols, abstract data types, or recursive protocols.

It does not attempt to emulate Go's full post-1.18 type-set language such as
tilde type terms, unions of concrete types, or pointer/value receiver rules.
}
\examples{
local({
  area <- S7::new_generic("area", "x")
  draw <- S7::new_generic("draw", "x")

  Circle <- S7::new_class(
    "Circle",
    properties = list(r = S7::class_double)
  )
  Rect <- S7::new_class(
    "Rect",
    properties = list(w = S7::class_double, h = S7::class_double)
  )

  S7::method(area, Circle) <- function(x) pi * x@r^2
  S7::method(draw, Circle) <- function(x) sprintf("circle(r = \%s)", x@r)
  S7::method(area, Rect) <- function(x) x@w * x@h

  Drawable <- new_interface("Drawable", generics = list(draw = draw))
  Shape <- new_interface("Shape", generics = list(area = area), parents = Drawable)

  implements(Circle, Shape)
  missing_requirements(Rect, Shape)
})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/trait.R
\name{trait_methods}
\alias{trait_methods}
\alias{impl_trait}
\alias{trait_report}
\alias{has_trait}
\alias{assert_trait}
\alias{trait_call}
\alias{trait_assoc_type}
\alias{trait_assoc_const}
\title{Inspect or use a Rust-like explicit trait}
\usage{
trait_methods(trait, inherited = TRUE)

impl_trait(
  trait,
  class,
  methods = list(),
  assoc_types = list(),
  assoc_consts = list(),
  replace = FALSE
)

trait_report(x, trait)

has_trait(x, trait)

assert_trait(x, trait, arg = deparse(substitute(x)))

trait_call(trait, method, x, ...)

trait_assoc_type(trait, x, name)

trait_assoc_const(trait, x, name)
}
\arguments{
\item{trait}{A trait created by \code{new_trait()}.}

\item{inherited}{Include inherited methods from supertraits?}

\item{class}{An S7 class or base class wrapper.}

\item{methods}{Named list of method implementations. Omitted trait methods
use their default implementation when one is available.}

\item{assoc_types}{Named list of associated type values.}

\item{assoc_consts}{Named list of associated constant values.}

\item{replace}{Replace an existing implementation record and silence warnings
about visible S7 methods?}

\item{x}{An object or class.}

\item{arg}{Name to use in error messages.}

\item{method}{Method name within the trait.}

\item{...}{Additional arguments passed to the S7 generic.}

\item{name}{Associated item name.}
}
\value{
\code{trait_methods()} returns a named list of \code{trait_method()} objects.
\code{impl_trait()} returns the stored implementation record, invisibly.
\code{trait_report()} returns a one-row data frame. \code{has_trait()} returns a
single logical value. \code{assert_trait()} returns \code{x}, unchanged.
\code{trait_call()} returns the result of the underlying S7 generic.
\code{trait_assoc_type()} and \code{trait_assoc_const()} return the stored associated
item value.
}
\description{
Inspect or use a Rust-like explicit trait
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/package.R
\docType{package}
\name{s7contract}
\alias{s7contract-package}
\alias{s7contract}
\title{s7contract: Contract Helpers for S7}
\description{
\code{s7contract} provides two experimental contract layers on top of S7:
}
\details{
\itemize{
\item Go-like structural interfaces defined by required generics.
\item Rust-like explicit traits with default methods and associated metadata.
}

The package keeps actual method dispatch inside ordinary S7 generics and uses
runtime checks to describe or assert conformance.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/sounkou-bioinfo/s7contract}
  \item \url{https://sounkou-bioinfo.github.io/s7contract/}
  \item Report bugs at \url{https://github.com/sounkou-bioinfo/s7contract/issues}
}

}
\author{
\strong{Maintainer}: Sounkou Mahamane Toure \email{sounkoutoure@gmail.com}

}
